// Copyright (C) 2016  University of Malta

// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

use std::fs;
use std::io;
use std::iter;

use std::io::Read;

use super::error::Error;

#[derive(Clone, PartialEq, Debug)]
pub struct SourceSet {
    set: Vec<Source>,
}

#[derive(Clone, PartialEq, Debug)]
struct Source {
    filename: String,
    contents: String,
}

#[derive(Copy, Clone, PartialEq, Debug)]
pub struct Span {
    source_id: usize,
    pub start: usize,
    pub end: usize,
    pub line: usize,
    pub column: usize,
}

#[derive(Copy, Clone, PartialEq, Debug)]
pub struct SourceSpans<'a> {
    index: usize,
    sources: &'a SourceSet,
}

impl<'a> SourceSpans<'a> {
    fn new(sources: &SourceSet) -> SourceSpans {
        SourceSpans {
            index: 0,
            sources: sources,
        }
    }
}

impl<'a> iter::Iterator for SourceSpans<'a> {
    type Item = Span;
    fn next(&mut self) -> Option<Span> {
        let i = self.index;
        match self.sources.set.get(i) {
            Some(source) => {
                self.index += 1;
                Some(Span {
                    source_id: i,
                    start: 0,
                    end: source.contents.len(),
                    line: 1,
                    column: 1,
                })
            }
            None => None,
        }
    }
}

impl SourceSet {
    pub fn new() -> SourceSet {
        SourceSet { set: Vec::new() }
    }

    pub fn read(&mut self, filename: &str) -> Result<(), Error> {
        self.set.push(Source::read_from(filename)?);
        Ok(())
    }

    pub fn source_spans(&self) -> SourceSpans {
        SourceSpans::new(self)
    }
}

impl Source {
    fn read_from(filename: &str) -> Result<Source, Error> {
        let file = fs::File::open(filename).map_err(|e| Error::io(filename, e))?;
        let mut reader = io::BufReader::new(file);
        let mut contents = String::new();
        reader.read_to_string(&mut contents).map_err(|e| Error::io(filename, e))?;
        Ok(Source {
            filename: filename.to_string(),
            contents: contents,
        })
    }
}

impl Span {
    pub fn get<'a>(&self, sources: &'a SourceSet) -> &'a str {
        &sources.set[self.source_id].contents[self.start..self.end]
    }

    pub fn get_whole<'a>(&self, sources: &'a SourceSet) -> &'a str {
        &sources.set[self.source_id].contents
    }

    pub fn filename<'a>(&self, sources: &'a SourceSet) -> &'a str {
        &sources.set[self.source_id].filename
    }

    pub fn err(&self, sources: &SourceSet, message: &str) -> Error {
        let found_str = format!("'{}'", self.get(sources));
        Error::parse(&format!("{}:{}:{} Found {}, {}",
                              sources.set[self.source_id].filename,
                              self.line,
                              self.column,
                              &found_str,
                              message))
    }

    pub fn skip_whitespace(&mut self, sources: &SourceSet) {
        let stream = &sources.set[self.source_id].contents;
        enum CommentState {
            Outside,
            Slash(usize), // position of the slash in the stream
            SlashSlash,
            SlashStar(usize), // nesting level
            SlashStarStar(usize), // nesting level
            SlashStarSlash(usize), // nesting level
        };
        let mut comment_state = CommentState::Outside;
        for (index, c) in stream[self.start..self.end].char_indices() {
            let (prev_line, prev_column) = (self.line, self.column);
            if c == '\n' {
                self.line += 1;
                self.column = 1;
            } else {
                self.column += 1;
            }
            comment_state = match comment_state {
                CommentState::Outside => {
                    match c {
                        '\n' | '\t' | '\r' | ' ' => comment_state,
                        '/' => CommentState::Slash(index),
                        _ => {
                            self.column -= 1;
                            self.start += index;
                            return;
                        }
                    }
                }
                CommentState::Slash(pos) => {
                    match c {
                        '/' => CommentState::SlashSlash,
                        '*' => CommentState::SlashStar(0),
                        _ => {
                            self.line = prev_line;
                            self.column = prev_column - 1;
                            self.start += pos;
                            return;
                        }
                    }
                }
                CommentState::SlashSlash => {
                    match c {
                        '\n' => CommentState::Outside,
                        _ => comment_state,
                    }
                }
                CommentState::SlashStar(nesting) => {
                    match c {
                        '/' => CommentState::SlashStarSlash(nesting),
                        '*' => CommentState::SlashStarStar(nesting),
                        _ => comment_state,
                    }
                }
                CommentState::SlashStarSlash(nesting) => {
                    match c {
                        '/' => comment_state,
                        '*' => CommentState::SlashStar(nesting + 1),
                        _ => CommentState::SlashStar(nesting),
                    }
                }
                CommentState::SlashStarStar(nesting) => {
                    match c {
                        '/' if nesting == 0 => CommentState::Outside,
                        '/' => CommentState::SlashStar(nesting - 1),
                        '*' => comment_state,
                        _ => CommentState::SlashStar(nesting),
                    }
                }
            }
        }
        if let CommentState::Slash(pos) = comment_state {
            self.column -= 1;
            self.start += pos;
        } else {
            self.start = self.end;
        }
    }
}
