// Copyright (C) 2016  University of Malta

// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

use std::clone;
use std::ffi;
use std::fmt;
use std::mem;
use std::ops;
use std::ptr;

use std::ops::{Add, AddAssign, Sub, SubAssign};
use std::ops::{Mul, MulAssign, Div, DivAssign};
use std::ops::{Shl, ShlAssign, Shr, ShrAssign};
use std::ops::Neg;
use std::cmp::{PartialEq, PartialOrd, Ordering};

use std::os::raw::c_int;

use super::gmp;
use super::mpfr;
use super::integer;
use super::rational;

use super::{Set, Get};
use super::{FromPrec, FromPrecRound};
use super::{SetRound, GetRound};
use super::{AddRound, SubRound, MulRound, DivRound};
use super::{ShlRound, ShrRound};
use super::{Pow, PowRound, PowAssign};
use super::{Integer, Rational};

pub struct Float(mpfr::__mpfr_struct);

pub fn raw(f: &Float) -> &mpfr::__mpfr_struct {
    &f.0
}

pub struct Exp(pub mpfr::mpfr_exp_t);
pub struct Prec(pub mpfr::mpfr_prec_t);

pub enum Round {
    Nearest,
    Zero,
    Up,
    Down,
    AwayFromZero,
}

pub enum Constant {
    Log2,
    Pi,
    Euler,
    Catalan,
}

pub enum Special {
    Zero,
    MinusZero,
    Infinity,
    MinusInfinity,
    Nan,
}

impl From<Round> for mpfr::mpfr_rnd_t {
    fn from(round: Round) -> mpfr::mpfr_rnd_t {
        match round {
            Round::Nearest => mpfr::mpfr_rnd_t::MPFR_RNDN,
            Round::Zero => mpfr::mpfr_rnd_t::MPFR_RNDZ,
            Round::Up => mpfr::mpfr_rnd_t::MPFR_RNDU,
            Round::Down => mpfr::mpfr_rnd_t::MPFR_RNDD,
            Round::AwayFromZero => mpfr::mpfr_rnd_t::MPFR_RNDA,
        }
    }
}

pub fn get_exp_range() -> (Exp, Exp) {
    unsafe { (Exp(mpfr::mpfr_get_emin()), Exp(mpfr::mpfr_get_emax())) }
}

impl Float {
    pub fn new(mut prec: Prec) -> Float {
        unsafe {
            const MPFR_PREC_MIN: mpfr::mpfr_prec_t = 2;
            if prec.0 < MPFR_PREC_MIN {
                prec.0 = MPFR_PREC_MIN;
            }
            let mut data: mpfr::__mpfr_struct = mem::uninitialized();
            mpfr::mpfr_init2(&mut data, prec.0);
            mpfr::mpfr_set_zero(&mut data, 0);
            Float(data)
        }
    }
}

impl ops::Drop for Float {
    fn drop(&mut self) {
        unsafe { mpfr::mpfr_clear(&mut self.0) };
    }
}

impl clone::Clone for Float {
    fn clone(&self) -> Float {
        let mut ret = Float::new(self.prec());
        ret.set(self);
        ret
    }

    fn clone_from(&mut self, source: &Float) {
        self.set_prec(source.prec());
        self.set(source);
    }
}

impl Float {
    pub fn prec(&self) -> Prec {
        Prec(unsafe { mpfr::mpfr_get_prec(&self.0) })
    }

    pub fn set_prec(&mut self, prec: Prec) {
        unsafe { mpfr::mpfr_set_prec(&mut self.0, prec.0) };
    }

    pub fn from_constant_round(prec: Prec, constant: Constant, round: Round) -> Float {
        let mut ret = Float::new(prec);
        ret.set_constant_round(constant, round);
        ret
    }

    pub fn from_constant(prec: Prec, constant: Constant) -> Float {
        Float::from_constant_round(prec, constant, Round::Nearest)
    }

    pub fn set_constant_round(&mut self, constant: Constant, round: Round) {
        unsafe {
            match constant {
                Constant::Log2 => mpfr::mpfr_const_log2(&mut self.0, round.into()),
                Constant::Pi => mpfr::mpfr_const_pi(&mut self.0, round.into()),
                Constant::Euler => mpfr::mpfr_const_euler(&mut self.0, round.into()),
                Constant::Catalan => mpfr::mpfr_const_catalan(&mut self.0, round.into()),
            };
        }
    }

    pub fn set_constant(&mut self, constant: Constant) {
        self.set_constant_round(constant, Round::Nearest);
    }

    pub fn set_special(&mut self, special: Special) {
        unsafe {
            match special {
                Special::Zero => mpfr::mpfr_set_zero(&mut self.0, 0),
                Special::MinusZero => mpfr::mpfr_set_zero(&mut self.0, -1),
                Special::Infinity => mpfr::mpfr_set_inf(&mut self.0, 0),
                Special::MinusInfinity => mpfr::mpfr_set_inf(&mut self.0, -1),
                Special::Nan => mpfr::mpfr_set_nan(&mut self.0),
            };
        }
    }
}

impl<T> FromPrec<T> for Float
    where Float: Set<T>
{
    fn from_prec(prec: Prec, t: T) -> Float {
        let mut ret = Float::new(prec);
        ret.set(t);
        ret
    }
}

impl<T> FromPrecRound<T> for Float
    where Float: SetRound<T>
{
    fn from_prec_round(prec: Prec, t: T, round: Round) -> Float {
        let mut ret = Float::new(prec);
        ret.set_round(t, round);
        ret
    }
}

macro_rules! set_other_ops{
    ($($t:ty)*) => {$(
        impl<'a> Set<&'a $t> for Float {
            fn set(&mut self, other: &'a $t) {
                self.set_round(other, Round::Nearest);
            }
        }

        impl Set<$t> for Float {
            fn set(&mut self, other: $t) {
                self.set_round(&other, Round::Nearest);
            }
        }

        impl SetRound<$t> for Float {
            fn set_round(&mut self, other: $t, round: Round) {
                self.set_round(&other, round);
            }
        }
    )*};
}

impl<'a> SetRound<&'a Float> for Float {
    fn set_round(&mut self, other: &'a Float, round: Round) {
        unsafe { mpfr::mpfr_set(&mut self.0, &other.0, round.into()) };
    }
}

impl<'a> SetRound<&'a Integer> for Float {
    fn set_round(&mut self, other: &'a Integer, round: Round) {
        let ptr = integer::raw(other) as *const gmp::__mpz_struct as *const mpfr::__mpz_struct;
        unsafe { mpfr::mpfr_set_z(&mut self.0, ptr, round.into()) };
    }
}

impl<'a> SetRound<&'a Rational> for Float {
    fn set_round(&mut self, other: &'a Rational, round: Round) {
        let ptr = rational::raw(other) as *const gmp::__mpq_struct as *const mpfr::__mpq_struct;
        unsafe { mpfr::mpfr_set_q(&mut self.0, ptr, round.into()) };
    }
}

set_other_ops!{ Float Integer Rational }

macro_rules! arith_op {
    ($imp:ident $method:ident,
     $imp_round:ident $method_round:ident,
     $imp_assign:ident $method_assign:ident,
     forward $t:ty, $func:path) => {
        impl $imp<$t> for Float {
            type Output = Float;
            fn $method(self, op: $t) -> Float {
                self.$method_round(op, Round::Nearest)
            }
        }

        impl $imp_round<$t> for Float {
            type Output = Float;
            fn $method_round(mut self, op: $t, round: Round) -> Float {
                unsafe { $func(&mut self.0, &self.0, op, round.into()) };
                self
            }
        }

        impl $imp_assign<$t> for Float {
            fn $method_assign(&mut self, op: $t) {
                unsafe { $func(&mut self.0, &self.0, op, Round::Nearest.into()) };
            }
        }
    };

    ($imp:ident $method:ident,
     $imp_round:ident $method_round:ident,
     $imp_assign:ident $method_assign:ident,
     $t:ty, $func:path, $func_from:path) => {
        arith_op!{ $imp $method, $imp_round $method_round, $imp_assign $method_assign,
                   forward $t, $func }

        impl $imp<Float> for $t {
            type Output = Float;
            fn $method(self, op: Float) -> Float {
                self.$method_round(op, Round::Nearest)
            }
        }

        impl<'a> $imp<&'a Float> for $t {
            type Output = Float;
            fn $method(self, op: &'a Float) -> Float {
                self.$method_round(op, Round::Nearest)
            }
        }

        impl $imp_round<Float> for $t {
            type Output = Float;
            fn $method_round(self, mut op: Float, round: Round) -> Float {
                unsafe { $func_from(&mut op.0, self, &op.0, round.into()) };
                op
            }
        }

        impl<'a> $imp_round<&'a Float> for $t {
            type Output = Float;
            fn $method_round(self, op: &'a Float, round: Round) -> Float {
                self.$method_round(op.clone(), round)
            }
        }
    };

    ($imp:ident $method:ident,
     $imp_round:ident $method_round:ident,
     $imp_assign:ident $method_assign:ident,
     $t:ty, $func:path) => {
        arith_op!{ $imp $method, $imp_round $method_round, $imp_assign $method_assign,
                   forward $t, $func }

        impl $imp<Float> for $t {
            type Output = Float;
            fn $method(self, op: Float) -> Float {
                self.$method_round(op, Round::Nearest)
            }
        }

        impl<'a> $imp<&'a Float> for $t {
            type Output = Float;
            fn $method(self, op: &'a Float) -> Float {
                self.$method_round(op, Round::Nearest)
            }
        }

        impl $imp_round<Float> for $t {
            type Output = Float;
            fn $method_round(self, op: Float, round: Round) -> Float {
                op.$method_round(self, round)
            }
        }

        impl<'a> $imp_round<&'a Float> for $t {
            type Output = Float;
            fn $method_round(self, op: &'a Float, round: Round) -> Float {
                self.$method_round(op.clone(), round)
            }
        }
    };

    ($imp:ident $method:ident,
     $imp_round:ident $method_round:ident,
     $imp_assign:ident $method_assign:ident,
     $func:path) => {
        impl<'a> $imp<&'a Float> for Float {
            type Output = Float;
            fn $method(self, op: &'a Float) -> Float {
                self.$method_round(op, Round::Nearest)
            }
        }

        impl $imp<Float> for Float {
            type Output = Float;
           fn $method(self, op: Float) -> Float {
                self.$method_round(op, Round::Nearest)
            }
        }

        impl<'a> $imp_round<&'a Float> for Float {
            type Output = Float;
            fn $method_round(mut self, op: &'a Float, round: Round) -> Float {
                unsafe { $func(&mut self.0, &self.0, &op.0, round.into()) };
                self
            }
        }

        impl $imp_round<Float> for Float {
            type Output = Float;
            fn $method_round(self, op: Float, round: Round) -> Float {
                self.$method_round(&op, round)
            }
        }

        impl<'a> $imp_assign<&'a Float> for Float {
            fn $method_assign(&mut self, op: &'a Float) {
                unsafe { $func(&mut self.0, &self.0, &op.0, Round::Nearest.into()) };
            }
        }

        impl $imp_assign<Float> for Float {
            fn $method_assign(&mut self, op: Float) {
                self.$method_assign(&op);
            }
        }
    }
}

arith_op!{ Add add, AddRound add_round, AddAssign add_assign, mpfr::mpfr_add }
arith_op!{ Sub sub, SubRound sub_round, SubAssign sub_assign, mpfr::mpfr_sub }
arith_op!{ Mul mul, MulRound mul_round, MulAssign mul_assign, mpfr::mpfr_mul }
arith_op!{ Div div, DivRound div_round, DivAssign div_assign, mpfr::mpfr_div }

macro_rules! sh_op {
    ($imp:ident $method:ident,
     $imp_round:ident $method_round:ident,
     $imp_assign:ident $method_assign:ident,
     $t:ty, $func:path) => {
        impl $imp<$t> for Float {
            type Output = Float;
            fn $method(self, op: $t) -> Float {
                self.$method_round(op, Round::Nearest)
            }
        }

        impl $imp_round<$t> for Float {
            type Output = Float;
            fn $method_round(mut self, op: $t, round: Round) -> Float {
                unsafe { $func(&mut self.0, &self.0, op, round.into()) };
                self
            }
        }

        impl $imp_assign<$t> for Float {
            fn $method_assign(&mut self, op: $t) {
                unsafe { $func(&mut self.0, &self.0, op, Round::Nearest.into()) };
            }
        }
    }
}

sh_op!{ Shl shl, ShlRound shl_round, ShlAssign shl_assign, u64, mpfr::mpfr_mul_2ui }
sh_op!{ Shr shr, ShrRound shr_round, ShrAssign shr_assign, u64, mpfr::mpfr_div_2ui }
sh_op!{ Shl shl, ShlRound shl_round, ShlAssign shl_assign, i64, mpfr::mpfr_mul_2si }
sh_op!{ Shr shr, ShrRound shr_round, ShrAssign shr_assign, i64, mpfr::mpfr_div_2si }

macro_rules! pow_other_ops {
    ($($t:ty)*) => {$(
        impl<'a> Pow<&'a $t> for Float {
            type Output = Float;
            fn pow(self, op: &'a $t) -> Float {
                self.pow_round(op, Round::Nearest)
            }
        }

        impl Pow<$t> for Float {
            type Output = Float;
            fn pow(self, op: $t) -> Float {
                self.pow_round(op, Round::Nearest)
            }
        }

        impl PowRound<$t> for Float {
            type Output = Float;
            fn pow_round(self, op: $t, round: Round) -> Float {
                self.pow_round(&op, round)
            }
        }

        impl PowAssign<$t> for Float {
            fn pow_assign(&mut self, op: $t) {
                self.pow_assign(&op);
            }
        }
    )*}
}

impl<'a> PowRound<&'a Float> for Float {
    type Output = Float;
    fn pow_round(mut self, op: &'a Float, round: Round) -> Float {
        unsafe { mpfr::mpfr_pow(&mut self.0, &self.0, &op.0, round.into()) };
        self
    }
}

impl<'a> PowAssign<&'a Float> for Float {
    fn pow_assign(&mut self, op: &'a Float) {
        unsafe { mpfr::mpfr_pow(&mut self.0, &self.0, &op.0, Round::Nearest.into()) };
    }
}

impl<'a> PowRound<&'a Integer> for Float {
    type Output = Float;
    fn pow_round(mut self, op: &'a Integer, round: Round) -> Float {
        let ptr = integer::raw(op) as *const gmp::__mpz_struct as *const mpfr::__mpz_struct;
        unsafe { mpfr::mpfr_pow_z(&mut self.0, &self.0, ptr, round.into()) };
        self
    }
}

impl<'a> PowAssign<&'a Integer> for Float {
    fn pow_assign(&mut self, op: &'a Integer) {
        let ptr = integer::raw(op) as *const gmp::__mpz_struct as *const mpfr::__mpz_struct;
        unsafe { mpfr::mpfr_pow_z(&mut self.0, &self.0, ptr, Round::Nearest.into()) };
    }
}

pow_other_ops!{ Float Integer }

macro_rules! conv_ops{
    ($t:ty, $set:path, $get:path,
     $add:path, $sub: path, $sub_from: path, $mul:path, $div: path, $div_from: path) => {
        impl Set<$t> for Float {
            fn set(&mut self, value: $t) {
                self.set_round(value, Round::Nearest);
            }
        }

        impl SetRound<$t> for Float {
            fn set_round(&mut self, value: $t, round: Round) {
                unsafe { $set(&mut self.0, value, round.into()) };
            }
        }

        impl Get<$t> for Float {
            fn get(&self) -> $t {
                self.get_round(Round::Nearest)
            }
        }

        impl GetRound<$t> for Float {
            fn get_round(&self, round: Round) -> $t {
                unsafe { $get(&self.0, round.into()) }
            }
        }

        arith_op!{ Add add, AddRound add_round, AddAssign add_assign, $t, $add }
        arith_op!{ Sub sub, SubRound sub_round, SubAssign sub_assign, $t, $sub, $sub_from }
        arith_op!{ Mul mul, MulRound mul_round, MulAssign mul_assign, $t, $mul }
        arith_op!{ Div div, DivRound div_round, DivAssign div_assign, $t, $div, $div_from }
    }
}

conv_ops!{ u64, mpfr::mpfr_set_ui, mpfr::mpfr_get_ui,
           mpfr::mpfr_add_ui, mpfr::mpfr_sub_ui, mpfr::mpfr_ui_sub,
           mpfr::mpfr_mul_ui, mpfr::mpfr_div_ui, mpfr::mpfr_ui_div }
conv_ops!{ i64, mpfr::mpfr_set_si, mpfr::mpfr_get_si,
           mpfr::mpfr_add_si, mpfr::mpfr_sub_si, mpfr::mpfr_si_sub,
           mpfr::mpfr_mul_si, mpfr::mpfr_div_si, mpfr::mpfr_si_div }
conv_ops!{ f64, mpfr::mpfr_set_d, mpfr::mpfr_get_d,
           mpfr::mpfr_add_d, mpfr::mpfr_sub_d, mpfr::mpfr_d_sub,
           mpfr::mpfr_mul_d, mpfr::mpfr_div_d, mpfr::mpfr_d_div }

impl Pow<u64> for Float {
    type Output = Float;
    fn pow(self, op: u64) -> Float {
        self.pow_round(op, Round::Nearest)
    }
}

impl Pow<Float> for u64 {
    type Output = Float;
    fn pow(self, op: Float) -> Float {
        self.pow_round(op, Round::Nearest)
    }
}

impl<'a> Pow<&'a Float> for u64 {
    type Output = Float;
    fn pow(self, op: &'a Float) -> Float {
        self.pow_round(op, Round::Nearest)
    }
}

impl PowRound<u64> for Float {
    type Output = Float;
    fn pow_round(mut self, op: u64, round: Round) -> Float {
        unsafe { mpfr::mpfr_pow_ui(&mut self.0, &self.0, op, round.into()) };
        self
    }
}

impl PowRound<Float> for u64 {
    type Output = Float;
    fn pow_round(self, mut op: Float, round: Round) -> Float {
        unsafe { mpfr::mpfr_ui_pow(&mut op.0, self, &op.0, round.into()) };
        op
    }
}

impl<'a> PowRound<&'a Float> for u64 {
    type Output = Float;
    fn pow_round(self, op: &'a Float, round: Round) -> Float {
        self.pow_round(op.clone(), round)
    }
}

impl PowAssign<u64> for Float {
    fn pow_assign(&mut self, op: u64) {
        unsafe { mpfr::mpfr_pow_ui(&mut self.0, &self.0, op, Round::Nearest.into()) };
    }
}

impl Pow<i64> for Float {
    type Output = Float;
    fn pow(self, op: i64) -> Float {
        self.pow_round(op, Round::Nearest)
    }
}

impl PowRound<i64> for Float {
    type Output = Float;
    fn pow_round(mut self, op: i64, round: Round) -> Float {
        unsafe { mpfr::mpfr_pow_si(&mut self.0, &self.0, op, round.into()) };
        self
    }
}

impl PowAssign<i64> for Float {
    fn pow_assign(&mut self, op: i64) {
        unsafe { mpfr::mpfr_pow_si(&mut self.0, &self.0, op, Round::Nearest.into()) };
    }
}

macro_rules! math_op1 {
    ($method:ident $method_round:ident, (u), $func:path
     $(, $param:ident: $t:ty)*) => {
        impl Float {
            pub fn $method(self, u: u64 $(, $param: $t)*) -> Float {
                self.$method_round(u, $($param, )* Round::Nearest)
            }

            pub fn $method_round(mut self, u: u64 $(, $param: $t)*, round: Round) -> Float {
                unsafe { $func(&mut self.0, u, $($param, )* round.into()) };
                self
            }
        }
    };

    ($method:ident $method_round:ident, $func:path
     $(, $param:ident: $t:ty)*) => {
        impl Float {
            pub fn $method(self $(, $param: $t)*) -> Float {
                self.$method_round($($param, )* Round::Nearest)
            }

            pub fn $method_round(mut self $(, $param: $t)*, round: Round) -> Float {
                unsafe { $func(&mut self.0, &self.0, $($param, )* round.into()) };
                self
            }
        }
    };

    ($method:ident $method_round:ident $method_buf:ident $method_round_buf:ident, $func:path
     $(, $param:ident: $t:ty)*) => {
        impl Float {
            pub fn $method(self $(, $param: $t)*) -> (Float, Float) {
                self.$method_round($($param, )* Round::Nearest)
            }

            pub fn $method_round(self $(, $param: $t)*, round: Round) -> (Float, Float) {
                let buf = Float::new(self.prec());
                self.$method_round_buf(buf, $($param, )* round)
            }

            pub fn $method_buf(self, buf: Float $(, $param: $t)*) -> (Float, Float) {
                self.$method_round_buf(buf, $($param, )* Round::Nearest)
            }

            pub fn $method_round_buf(mut self, mut buf: Float $(, $param: $t)*, round: Round)
                                     -> (Float, Float) {
                unsafe { $func(&mut self.0, &mut buf.0, &self.0, $($param, )* round.into()) };
                (self, buf)
            }
        }
    };
}

macro_rules! math_op2 {
    ($method:ident $method_round:ident, $func:path
     $(, $param:ident: $t:ty)*) => {
        impl Float {
            pub fn $method(self, other: &Float $(, $param: $t)*) -> Float {
                self.$method_round(other, $($param, )* Round::Nearest)
            }

            pub fn $method_round(mut self, other: &Float $(, $param: $t)*, round: Round) -> Float {
                unsafe { $func(&mut self.0, &self.0, &other.0, $($param, )* round.into()) };
                self
            }
        }
    }
}

math_op1!{ sqrt sqrt_round, mpfr::mpfr_sqrt }
math_op1!{ rec_sqrt rec_sqrt_round, mpfr::mpfr_rec_sqrt }
math_op1!{ cbrt cbrt_round, mpfr::mpfr_cbrt }
math_op1!{ root root_round, mpfr::mpfr_root, k: u64 }
math_op1!{ abs abs_round, mpfr::mpfr_abs }
math_op1!{ log log_round, mpfr::mpfr_log }
math_op1!{ log2 log2_round, mpfr::mpfr_log2 }
math_op1!{ log10 log10_round, mpfr::mpfr_log10 }
math_op1!{ exp exp_round, mpfr::mpfr_exp }
math_op1!{ exp2 exp2_round, mpfr::mpfr_exp2 }
math_op1!{ exp10 exp10_round, mpfr::mpfr_exp10 }
math_op1!{ cos cos_round, mpfr::mpfr_cos }
math_op1!{ sin sin_round, mpfr::mpfr_sin }
math_op1!{ tan tan_round, mpfr::mpfr_tan }
math_op1!{ sin_cos sin_cos_round sin_cos_buf sin_cos_round_buf, mpfr::mpfr_sin_cos }
math_op1!{ sec sec_round, mpfr::mpfr_sec }
math_op1!{ csc csc_round, mpfr::mpfr_csc }
math_op1!{ cot cot_round, mpfr::mpfr_cot }
math_op1!{ acos acos_round, mpfr::mpfr_acos }
math_op1!{ asin asin_round, mpfr::mpfr_asin }
math_op1!{ atan atan_round, mpfr::mpfr_atan }
math_op2!{ atan2 atan2_round, mpfr::mpfr_atan2 }
math_op1!{ cosh cosh_round, mpfr::mpfr_cosh }
math_op1!{ sinh sinh_round, mpfr::mpfr_sinh }
math_op1!{ tanh tanh_round, mpfr::mpfr_tanh }
math_op1!{ sinh_cosh sin_cosh_round sinh_cosh_buf sinh_cosh_round_buf, mpfr::mpfr_sinh_cosh }
math_op1!{ sech sech_round, mpfr::mpfr_sech }
math_op1!{ csch csch_round, mpfr::mpfr_csch }
math_op1!{ coth coth_round, mpfr::mpfr_coth }
math_op1!{ acosh acosh_round, mpfr::mpfr_acosh }
math_op1!{ asinh asinh_round, mpfr::mpfr_asinh }
math_op1!{ atanh atanh_round, mpfr::mpfr_atanh }
math_op1!{ log1p log1p_round, mpfr::mpfr_log1p }
math_op1!{ expm1 expm1_round, mpfr::mpfr_expm1 }
math_op1!{ eint eint_round, mpfr::mpfr_eint }
math_op1!{ li2 li2_round, mpfr::mpfr_li2 }
math_op1!{ gamma gamma_round, mpfr::mpfr_gamma }
math_op1!{ lngamma lngamma_round, mpfr::mpfr_lngamma }
math_op1!{ digamma digamma_round, mpfr::mpfr_digamma }
math_op1!{ zeta zeta_round, mpfr::mpfr_zeta }
math_op1!{ set_zeta set_zeta_round, (u), mpfr::mpfr_zeta_ui }
math_op1!{ erf erf_round, mpfr::mpfr_erf }
math_op1!{ erfc erfc_round, mpfr::mpfr_erfc }
math_op1!{ j0 j0_round, mpfr::mpfr_j0 }
math_op1!{ j1 j1_round, mpfr::mpfr_j1 }
math_op1!{ jn jn_round, jn, n: i64 }
math_op1!{ y0 y0_round, mpfr::mpfr_y0 }
math_op1!{ y1 y1_round, mpfr::mpfr_y1 }
math_op1!{ yn yn_round, yn, n: i64 }
math_op2!{ agm agm_round, mpfr::mpfr_agm }
math_op2!{ hypot hypot_round, mpfr::mpfr_hypot }
math_op1!{ ai ai_round, mpfr::mpfr_ai }
math_op1!{ ceil ceil_round, mpfr::mpfr_rint_ceil }
math_op1!{ floor floor_round, mpfr::mpfr_rint_floor }
math_op1!{ round round_round, mpfr::mpfr_rint_round }
math_op1!{ trunc trunc_round, mpfr::mpfr_rint_trunc }

unsafe fn jn(rop: mpfr::mpfr_ptr, op: mpfr::mpfr_srcptr, n: i64, rnd: mpfr::mpfr_rnd_t) -> c_int {
    mpfr::mpfr_jn(rop, n, op, rnd)
}

unsafe fn yn(rop: mpfr::mpfr_ptr, op: mpfr::mpfr_srcptr, n: i64, rnd: mpfr::mpfr_rnd_t) -> c_int {
    mpfr::mpfr_yn(rop, n, op, rnd)
}

impl Float {
    pub fn neg_round(mut self, round: Round) -> Float {
        unsafe { mpfr::mpfr_neg(&mut self.0, &self.0, round.into()) };
        self
    }
}

impl Neg for Float {
    type Output = Float;
    fn neg(self) -> Float {
        self.neg_round(Round::Nearest)
    }
}

impl Float {
    fn is_integer(&self) -> bool {
        unsafe { mpfr::mpfr_integer_p(&self.0) != 0 }
    }

    fn is_nan(&self) -> bool {
        unsafe { mpfr::mpfr_nan_p(&self.0) != 0 }
    }

    fn is_inf(&self) -> bool {
        unsafe { mpfr::mpfr_inf_p(&self.0) != 0 }
    }

    fn is_number(&self) -> bool {
        unsafe { mpfr::mpfr_number_p(&self.0) != 0 }
    }

    fn is_zero(&self) -> bool {
        unsafe { mpfr::mpfr_zero_p(&self.0) != 0 }
    }

    fn is_regular(&self) -> bool {
        unsafe { mpfr::mpfr_regular_p(&self.0) != 0 }
    }

    fn sgn(&self) -> Ordering {
        let ord = unsafe { mpfr::mpfr_sgn(&self.0) };
        if ord < 0 {
            Ordering::Less
        } else if ord > 0 {
            Ordering::Greater
        } else {
            Ordering::Equal
        }
    }
}

impl PartialEq for Float {
    fn eq(&self, other: &Float) -> bool {
        unsafe { mpfr::mpfr_equal_p(&self.0, &other.0) != 0 }
    }
}

impl PartialOrd for Float {
    fn partial_cmp(&self, other: &Float) -> Option<Ordering> {
        if unsafe { mpfr::mpfr_unordered_p(&self.0, &other.0) } != 0 {
            return None;
        }
        let ord = unsafe { mpfr::mpfr_cmp(&self.0, &other.0) };
        if ord < 0 {
            Some(Ordering::Less)
        } else if ord > 0 {
            Some(Ordering::Greater)
        } else {
            Some(Ordering::Equal)
        }
    }

    fn lt(&self, other: &Float) -> bool {
        unsafe { mpfr::mpfr_less_p(&self.0, &other.0) != 0 }
    }

    fn le(&self, other: &Float) -> bool {
        unsafe { mpfr::mpfr_lessequal_p(&self.0, &other.0) != 0 }
    }

    fn gt(&self, other: &Float) -> bool {
        unsafe { mpfr::mpfr_greater_p(&self.0, &other.0) != 0 }
    }

    fn ge(&self, other: &Float) -> bool {
        unsafe { mpfr::mpfr_greaterequal_p(&self.0, &other.0) != 0 }
    }
}

impl Float {
    fn get_exp(&self) -> Option<Exp> {
        if self.is_regular() {
            Some(Exp(unsafe { mpfr::mpfr_get_exp(&self.0) }))
        } else {
            None
        }
    }

    fn get_sign(&self) -> bool {
        unsafe { mpfr::mpfr_signbit(&self.0) != 0 }
    }
}

impl Float {
    pub fn to_string(&self, base: c_int, round: Round) -> String {
        self.make_string(base, round, false)
    }

    fn make_string(&self, base: c_int, round: Round, to_upper: bool) -> String {
        let mut buf = String::new();
        let mut exp: mpfr::mpfr_exp_t;
        let s;
        let cstr;
        unsafe {
            exp = mem::uninitialized();
            s = mpfr::mpfr_get_str(ptr::null_mut(), &mut exp, base, 0, &self.0, round.into());
            assert!(!s.is_null());
            cstr = ffi::CStr::from_ptr(s);
        }
        let mut chars = cstr.to_str().unwrap().chars();
        let mut c = chars.next();
        if c == Some('-') {
            buf.push('-');
            c = chars.next();
        }
        let mut special = true;
        if let Some(x) = c {
            match x {
                '0'...'9' | 'a'...'z' | 'A'...'Z' => {
                    buf.push_str("0.");
                    special = false;
                }
                _ => {}
            };
            buf.push(x);
        }
        for c in chars {
            buf.push(c);
        }
        unsafe {
            mpfr::mpfr_free_str(s);
        }
        if !special {
            buf.push_str(&format!("*{}^{}", base, exp));
        }
        if !special && to_upper {
            buf = buf.to_uppercase();
        }
        buf
    }
}

impl fmt::Display for Float {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        write!(f, "{}", self.to_string(10, Round::Nearest))
    }
}
impl fmt::Debug for Float {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        write!(f,
               "{} (hex {})",
               self.make_string(10, Round::Nearest, false),
               self.make_string(16, Round::Nearest, false))
    }
}

impl fmt::Binary for Float {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        write!(f, "{}", self.make_string(2, Round::Nearest, false))
    }
}

impl fmt::Octal for Float {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        write!(f, "{}", self.make_string(8, Round::Nearest, false))
    }
}

impl fmt::LowerHex for Float {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        write!(f, "{}", self.make_string(16, Round::Nearest, false))
    }
}

impl fmt::UpperHex for Float {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        write!(f, "{}", self.make_string(16, Round::Nearest, true))
    }
}
