// Copyright (C) 2016  University of Malta

// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

use std::clone;
use std::default;
use std::ffi;
use std::fmt;
use std::mem;
use std::ops;
use std::ptr;

use std::ops::{Add, AddAssign, Sub, SubAssign};
use std::ops::{Mul, MulAssign, Div, DivAssign, Rem, RemAssign};
use std::ops::{Shl, ShlAssign, Shr, ShrAssign};
use std::ops::Neg;
use std::cmp::{Eq, Ord, PartialEq, PartialOrd, Ordering};
use std::ops::{BitAnd, BitAndAssign, BitOr, BitOrAssign};
use std::ops::{BitXor, BitXorAssign, Not};

use std::os::raw::c_int;
use std::os::raw::c_void;

use super::float;
use super::gmp;
use super::mpfr;
use super::rational;

use super::{Set, Get};
use super::{Rational, Float};
use super::Round;
use super::{Pow, PowAssign};

pub struct Integer(gmp::__mpz_struct);

pub fn raw(z: &Integer) -> &gmp::__mpz_struct {
    &z.0
}

impl Integer {
    pub fn new() -> Integer {
        unsafe {
            let mut data: gmp::__mpz_struct = mem::uninitialized();
            gmp::__gmpz_init(&mut data);
            Integer(data)
        }
    }
}

impl ops::Drop for Integer {
    fn drop(&mut self) {
        unsafe { gmp::__gmpz_clear(&mut self.0) };
    }
}

impl default::Default for Integer {
    fn default() -> Integer {
        Integer::new()
    }
}

impl clone::Clone for Integer {
    fn clone(&self) -> Integer {
        let mut ret = Integer::new();
        ret.set(self);
        ret
    }

    fn clone_from(&mut self, source: &Integer) {
        self.set(source);
    }
}

macro_rules! from_set{
    ($($t:ty),*) => {$(
        impl<'a> From<$t> for Integer
        {
            fn from(t: $t) -> Integer {
                let mut ret = Integer::new();
                ret.set(t);
                ret
            }
        }
    )*}
}

from_set!{ Rational, Float, u64, i64, f64, &'a Integer, &'a Rational, &'a Float }

macro_rules! set_other_ops{
    ($($t:ty)*) => {$(
        impl Set<$t> for Integer {
            fn set(&mut self, other: $t) {
                self.set(&other);
            }
        }
    )*}
}

impl<'a> Set<&'a Integer> for Integer {
    fn set(&mut self, other: &'a Integer) {
        unsafe { gmp::__gmpz_set(&mut self.0, &other.0) };
    }
}

impl<'a> Set<&'a Rational> for Integer {
    fn set(&mut self, value: &'a Rational) {
        unsafe { gmp::__gmpz_set_q(&mut self.0, rational::raw(value)) };
    }
}

impl<'a> Set<&'a Float> for Integer {
    fn set(&mut self, value: &'a Float) {
        let ptr = &mut self.0 as *mut gmp::__mpz_struct as *mut mpfr::__mpz_struct;
        unsafe { mpfr::mpfr_get_z(ptr, float::raw(value), Round::Zero.into()) };
    }
}

set_other_ops!{ Integer Rational Float }

impl Set<u64> for Integer {
    fn set(&mut self, value: u64) {
        unsafe { gmp::__gmpz_set_ui(&mut self.0, value) };
    }
}

impl Set<i64> for Integer {
    fn set(&mut self, value: i64) {
        unsafe { gmp::__gmpz_set_si(&mut self.0, value) };
    }
}

impl Set<f64> for Integer {
    fn set(&mut self, value: f64) {
        unsafe { gmp::__gmpz_set_d(&mut self.0, value) };
    }
}

impl Get<u64> for Integer {
    fn get(&self) -> u64 {
        unsafe { gmp::__gmpz_get_ui(&self.0) }
    }
}

impl Get<i64> for Integer {
    fn get(&self) -> i64 {
        unsafe { gmp::__gmpz_get_si(&self.0) }
    }
}

impl Get<f64> for Integer {
    fn get(&self) -> f64 {
        unsafe { gmp::__gmpz_get_d(&self.0) }
    }
}

macro_rules! arith_op {
    ($imp:ident $method:ident,
     $imp_assign:ident $method_assign:ident,
     forward $t:ty, $func:path) => {
        impl<'a> $imp<$t> for Integer {
            type Output = Integer;
            fn $method(mut self, op: $t) -> Integer {
                unsafe { $func(&mut self.0, &self.0, op) };
                self
            }
        }

        impl $imp_assign<$t> for Integer {
            fn $method_assign(&mut self, op: $t) {
                unsafe { $func(&mut self.0, &self.0, op) };
            }
        }
    };

    ($imp:ident $method:ident,
     $imp_assign:ident $method_assign:ident,
     $t:ty, $func:path, $func_from:path) => {
        arith_op!{ $imp $method, $imp_assign $method_assign, forward $t, $func }

        impl $imp<Integer> for $t {
            type Output = Integer;
            fn $method(self, mut op: Integer) -> Integer {
                unsafe { $func_from(&mut op.0, self, &op.0) };
                op
            }
        }

        impl<'a> $imp<&'a Integer> for $t {
            type Output = Integer;
            fn $method(self, op: &'a Integer) -> Integer {
                self.$method(op.clone())
            }
        }
    };

    ($imp:ident $method:ident,
     $imp_assign:ident $method_assign:ident,
     $t:ty, $func:path) => {
        arith_op!{ $imp $method, $imp_assign $method_assign, forward $t, $func }

        impl $imp<Integer> for $t {
            type Output = Integer;
            fn $method(self, op: Integer) -> Integer {
                op.$method(self)
            }
        }

        impl<'a> $imp<&'a Integer> for $t {
            type Output = Integer;
            fn $method(self, op: &'a Integer) -> Integer {
                self.$method(op.clone())
            }
        }
    };

    ($imp:ident $method:ident,
     $imp_assign:ident $method_assign:ident,
     $func:path) => {
        impl<'a> $imp<&'a Integer> for Integer {
            type Output = Integer;
            fn $method(mut self, op: &'a Integer) -> Integer {
                unsafe { $func(&mut self.0, &self.0, &op.0) };
                self
            }
        }

        impl $imp<Integer> for Integer {
            type Output = Integer;
            fn $method(self, op: Integer) -> Integer {
                self.$method(&op)
            }
        }

        impl<'a> $imp_assign<&'a Integer> for Integer {
            fn $method_assign(&mut self, op: &'a Integer) {
                unsafe { $func(&mut self.0, &self.0, &op.0) };
            }
        }

        impl $imp_assign<Integer> for Integer {
            fn $method_assign(&mut self, op: Integer) {
                self.$method_assign(&op);
            }
        }
    }
}

arith_op!{ Add add, AddAssign add_assign, gmp::__gmpz_add }
arith_op!{ Sub sub, SubAssign sub_assign, gmp::__gmpz_sub }
arith_op!{ Mul mul, MulAssign mul_assign, gmp::__gmpz_mul }
arith_op!{ Div div, DivAssign div_assign, gmp::__gmpz_tdiv_q }
arith_op!{ Rem rem, RemAssign rem_assign, gmp::__gmpz_tdiv_r }
arith_op!{ Add add, AddAssign add_assign, u64, gmp::__gmpz_add_ui }
arith_op!{ Sub sub, SubAssign sub_assign, u64, gmp::__gmpz_sub_ui, gmp::__gmpz_ui_sub }
arith_op!{ Mul mul, MulAssign mul_assign, u64, gmp::__gmpz_mul_ui }
arith_op!{ Mul mul, MulAssign mul_assign, i64, gmp::__gmpz_mul_si }
arith_op!{ Div div, DivAssign div_assign, forward u64, gmp::__gmpz_tdiv_q_ui }
arith_op!{ Rem rem, RemAssign rem_assign, forward u64, gmp::__gmpz_tdiv_r_ui }
arith_op!{ Shl shl, ShlAssign shl_assign, forward u64, gmp::__gmpz_mul_2exp }
arith_op!{ Shr shr, ShrAssign shr_assign, forward u64, gmp::__gmpz_fdiv_q_2exp }
arith_op!{ Pow pow, PowAssign pow_assign, forward u64, gmp::__gmpz_pow_ui }

impl Neg for Integer {
    type Output = Integer;
    fn neg(mut self) -> Integer {
        unsafe { gmp::__gmpz_neg(&mut self.0, &self.0) };
        self
    }
}

impl Integer {
    fn div_rem(mut self, mut divisor: Integer) -> (Integer, Integer) {
        unsafe { gmp::__gmpz_tdiv_qr(&mut self.0, &mut divisor.0, &self.0, &divisor.0) };
        (self, divisor)
    }

    fn abs(mut self) -> Integer {
        unsafe { gmp::__gmpz_abs(&mut self.0, &self.0) };
        self
    }

    fn root(mut self, n: u64) -> Integer {
        unsafe { gmp::__gmpz_root(&mut self.0, &self.0, n) };
        self
    }

    fn root_rem(mut self, mut rem_buf: Integer, n: u64) -> (Integer, Integer) {
        unsafe { gmp::__gmpz_rootrem(&mut self.0, &mut rem_buf.0, &self.0, n) };
        (self, rem_buf)
    }

    fn sqrt(mut self) -> Integer {
        unsafe { gmp::__gmpz_sqrt(&mut self.0, &self.0) };
        self
    }

    fn sqrt_rem(mut self, mut rem_buf: Integer) -> (Integer, Integer) {
        unsafe { gmp::__gmpz_sqrtrem(&mut self.0, &mut rem_buf.0, &self.0) };
        (self, rem_buf)
    }

    fn is_perfect_power(&self) -> bool {
        unsafe { gmp::__gmpz_perfect_power_p(&self.0) != 0 }
    }

    fn is_perfect_square(&self) -> bool {
        unsafe { gmp::__gmpz_perfect_square_p(&self.0) != 0 }
    }
}

impl Eq for Integer {}

impl Ord for Integer {
    fn cmp(&self, other: &Integer) -> Ordering {
        let ord = unsafe { gmp::__gmpz_cmp(&self.0, &other.0) };
        ord.cmp(&0)
    }
}

impl PartialEq for Integer {
    fn eq(&self, other: &Integer) -> bool {
        self.cmp(other) == Ordering::Equal
    }
}

impl PartialOrd for Integer {
    fn partial_cmp(&self, other: &Integer) -> Option<Ordering> {
        Some(self.cmp(other))
    }
}

impl PartialEq<f64> for Integer {
    fn eq(&self, other: &f64) -> bool {
        self.partial_cmp(other) == Some(Ordering::Equal)
    }
}

impl PartialOrd<f64> for Integer {
    fn partial_cmp(&self, other: &f64) -> Option<Ordering> {
        if other.is_nan() {
            None
        } else {
            let ord = unsafe { gmp::__gmpz_cmp_d(&self.0, *other) };
            Some(ord.cmp(&0))
        }
    }
}

impl PartialEq<Integer> for f64 {
    fn eq(&self, other: &Integer) -> bool {
        other.eq(self)
    }
}

impl PartialOrd<Integer> for f64 {
    fn partial_cmp(&self, other: &Integer) -> Option<Ordering> {
        match other.partial_cmp(self) {
            None => None,
            Some(x) => Some(x.reverse()),
        }
    }
}

impl PartialEq<i64> for Integer {
    fn eq(&self, other: &i64) -> bool {
        self.partial_cmp(other) == Some(Ordering::Equal)
    }
}

impl PartialOrd<i64> for Integer {
    fn partial_cmp(&self, other: &i64) -> Option<Ordering> {
        let ord = unsafe { gmp::__gmpz_cmp_si(&self.0, *other) };
        Some(ord.cmp(&0))
    }
}

impl PartialEq<Integer> for i64 {
    fn eq(&self, other: &Integer) -> bool {
        other.eq(self)
    }
}

impl PartialOrd<Integer> for i64 {
    fn partial_cmp(&self, other: &Integer) -> Option<Ordering> {
        match other.partial_cmp(self) {
            None => None,
            Some(x) => Some(x.reverse()),
        }
    }
}

impl PartialEq<u64> for Integer {
    fn eq(&self, other: &u64) -> bool {
        self.partial_cmp(other) == Some(Ordering::Equal)
    }
}

impl PartialOrd<u64> for Integer {
    fn partial_cmp(&self, other: &u64) -> Option<Ordering> {
        let ord = unsafe { gmp::__gmpz_cmp_ui(&self.0, *other) };
        Some(ord.cmp(&0))
    }
}

impl PartialEq<Integer> for u64 {
    fn eq(&self, other: &Integer) -> bool {
        other.eq(self)
    }
}

impl PartialOrd<Integer> for u64 {
    fn partial_cmp(&self, other: &Integer) -> Option<Ordering> {
        match other.partial_cmp(self) {
            None => None,
            Some(x) => Some(x.reverse()),
        }
    }
}

impl Integer {
    fn sgn(&self) -> Ordering {
        if (self.0)._mp_size < 0 {
            Ordering::Less
        } else if (self.0)._mp_size > 0 {
            Ordering::Greater
        } else {
            Ordering::Equal
        }
    }
}

impl<'a> BitAnd<&'a Integer> for Integer {
    type Output = Integer;
    fn bitand(mut self, op: &'a Integer) -> Integer {
        unsafe { gmp::__gmpz_and(&mut self.0, &self.0, &op.0) };
        self
    }
}

impl BitAnd for Integer {
    type Output = Integer;
    fn bitand(self, op: Integer) -> Integer {
        self.bitand(&op)
    }
}

impl<'a> BitAndAssign<&'a Integer> for Integer {
    fn bitand_assign(&mut self, op: &'a Integer) {
        unsafe { gmp::__gmpz_and(&mut self.0, &self.0, &op.0) };
    }
}

impl BitAndAssign<Integer> for Integer {
    fn bitand_assign(&mut self, op: Integer) {
        self.add_assign(&op);
    }
}

impl<'a> BitOr<&'a Integer> for Integer {
    type Output = Integer;
    fn bitor(mut self, op: &'a Integer) -> Integer {
        unsafe { gmp::__gmpz_ior(&mut self.0, &self.0, &op.0) };
        self
    }
}

impl BitOr for Integer {
    type Output = Integer;
    fn bitor(self, op: Integer) -> Integer {
        self.bitor(&op)
    }
}

impl<'a> BitOrAssign<&'a Integer> for Integer {
    fn bitor_assign(&mut self, op: &'a Integer) {
        unsafe { gmp::__gmpz_ior(&mut self.0, &self.0, &op.0) };
    }
}

impl BitOrAssign<Integer> for Integer {
    fn bitor_assign(&mut self, op: Integer) {
        self.add_assign(&op);
    }
}

impl<'a> BitXor<&'a Integer> for Integer {
    type Output = Integer;
    fn bitxor(mut self, op: &'a Integer) -> Integer {
        unsafe { gmp::__gmpz_xor(&mut self.0, &self.0, &op.0) };
        self
    }
}

impl BitXor for Integer {
    type Output = Integer;
    fn bitxor(self, op: Integer) -> Integer {
        self.bitxor(&op)
    }
}

impl<'a> BitXorAssign<&'a Integer> for Integer {
    fn bitxor_assign(&mut self, op: &'a Integer) {
        unsafe { gmp::__gmpz_xor(&mut self.0, &self.0, &op.0) };
    }
}

impl BitXorAssign<Integer> for Integer {
    fn bitxor_assign(&mut self, op: Integer) {
        self.add_assign(&op);
    }
}

impl Not for Integer {
    type Output = Integer;
    fn not(mut self) -> Integer {
        unsafe { gmp::__gmpz_com(&mut self.0, &self.0) };
        self
    }
}

impl Integer {
    pub fn size_in_bits(&self) -> u64 {
        unsafe { gmp::__gmpz_sizeinbase(&self.0, 2) as u64 }
    }

    pub fn popcount(&self) -> u64 {
        unsafe { gmp::__gmpz_popcount(&self.0) }
    }

    pub fn hamdist(&self, op: &Integer) -> u64 {
        unsafe { gmp::__gmpz_hamdist(&self.0, &op.0) }
    }

    pub fn find_0(&self, start: u64) -> u64 {
        unsafe { gmp::__gmpz_scan0(&self.0, start) }
    }

    pub fn find_1(&self, start: u64) -> u64 {
        unsafe { gmp::__gmpz_scan1(&self.0, start) }
    }

    pub fn set_bit(&mut self, index: u64, value: bool) {
        unsafe {
            if value {
                gmp::__gmpz_setbit(&mut self.0, index);
            } else {
                gmp::__gmpz_clrbit(&mut self.0, index);
            }
        }
    }

    pub fn get_bit(&self, index: u64) -> bool {
        unsafe { gmp::__gmpz_tstbit(&self.0, index) != 0 }
    }

    pub fn invert_bit(&mut self, index: u64) {
        unsafe { gmp::__gmpz_combit(&mut self.0, index) };
    }
}

impl Integer {
    pub fn to_string(&self, base: c_int) -> String {
        self.make_string(base, false)
    }

    fn make_string(&self, base: c_int, to_upper: bool) -> String {
        let s;
        let cstr;
        unsafe {
            s = gmp::__gmpz_get_str(ptr::null_mut(), base, &self.0);
            assert!(!s.is_null());
            cstr = ffi::CStr::from_ptr(s);
        }
        let mut buf = cstr.to_str().unwrap().to_string();
        unsafe {
            let mut free = None;
            gmp::__gmp_get_memory_functions(ptr::null_mut(), ptr::null_mut(), &mut free);
            let free = free.unwrap();
            let free_len = cstr.to_bytes().len() + 1;
            free(s as *mut c_void, free_len);
        }
        if to_upper {
            buf = buf.to_uppercase();
        }
        buf
    }
}

impl fmt::Display for Integer {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        write!(f, "{}", self.to_string(10))
    }
}

impl fmt::Debug for Integer {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        write!(f, "0x{}", self.make_string(16, false))
    }
}

impl fmt::Binary for Integer {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        let prefix = if f.alternate() { "0b" } else { "" };
        write!(f, "{}{}", prefix, self.make_string(2, false))
    }
}

impl fmt::Octal for Integer {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        let prefix = if f.alternate() { "0o" } else { "" };
        write!(f, "{}{}", prefix, self.make_string(8, false))
    }
}

impl fmt::LowerHex for Integer {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        let prefix = if f.alternate() { "0x" } else { "" };
        write!(f, "{}{}", prefix, self.make_string(16, false))
    }
}

impl fmt::UpperHex for Integer {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        let prefix = if f.alternate() { "0X" } else { "" };
        write!(f, "{}{}", prefix, self.make_string(16, true))
    }
}
