// Copyright (C) 2016  University of Malta

// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

use std::clone;
use std::default;
use std::ffi;
use std::fmt;
use std::mem;
use std::ops;
use std::ptr;

use std::ops::{Add, AddAssign, Sub, SubAssign};
use std::ops::{Mul, MulAssign, Div, DivAssign};
use std::ops::{Shl, ShlAssign, Shr, ShrAssign};
use std::ops::Neg;
use std::cmp::{Eq, Ord, PartialEq, PartialOrd, Ordering};

use std::os::raw::c_int;
use std::os::raw::c_void;

use super::gmp;
use super::integer;

use super::{Set, Get};
use super::Integer;

pub struct Rational(gmp::__mpq_struct);

pub fn raw(q: &Rational) -> &gmp::__mpq_struct {
    &q.0
}

impl Rational {
    pub fn new() -> Rational {
        unsafe {
            let mut data: gmp::__mpq_struct = mem::uninitialized();
            gmp::__gmpq_init(&mut data);
            Rational(data)
        }
    }
}

impl ops::Drop for Rational {
    fn drop(&mut self) {
        unsafe { gmp::__gmpq_clear(&mut self.0) };
    }
}

impl default::Default for Rational {
    fn default() -> Rational {
        Rational::new()
    }
}

impl clone::Clone for Rational {
    fn clone(&self) -> Rational {
        let mut ret = Rational::new();
        ret.set(self);
        ret
    }

    fn clone_from(&mut self, source: &Rational) {
        self.set(source);
    }
}

macro_rules! from_set{
    ($($t:ty),*) => {$(
        impl<'a> From<$t> for Rational
        {
            fn from(t: $t) -> Rational {
                let mut ret = Rational::new();
                ret.set(t);
                ret
            }
        }
    )*}
}

from_set!{ Integer, (u64, u64), (i64, u64), &'a Integer, &'a Rational }

macro_rules! set_other_ops{
    ($($t:ty)*) => {$(
        impl<'a> Set<$t> for Rational {
            fn set(&mut self, other: $t) {
                self.set(&other);
            }
        }
    )*}
}

impl<'a> Set<&'a Rational> for Rational {
    fn set(&mut self, other: &'a Rational) {
        unsafe { gmp::__gmpq_set(&mut self.0, &other.0) };
    }
}

impl<'a> Set<&'a Integer> for Rational {
    fn set(&mut self, value: &'a Integer) {
        unsafe { gmp::__gmpq_set_z(&mut self.0, integer::raw(value)) };
    }
}

set_other_ops!{ Rational Integer }

impl Set<(u64, u64)> for Rational {
    fn set(&mut self, frac: (u64, u64)) {
        unsafe { gmp::__gmpq_set_ui(&mut self.0, frac.0, frac.1) };
    }
}

impl Set<(i64, u64)> for Rational {
    fn set(&mut self, frac: (i64, u64)) {
        unsafe { gmp::__gmpq_set_si(&mut self.0, frac.0, frac.1) };
    }
}

impl Get<f64> for Rational {
    fn get(&self) -> f64 {
        unsafe { gmp::__gmpq_get_d(&self.0) }
    }
}

impl Rational {
    pub fn set_num(&mut self, num: &Integer) {
        unsafe { gmp::__gmpq_set_num(&mut self.0, integer::raw(num)) };
    }

    pub fn set_den(&mut self, den: &Integer) {
        unsafe { gmp::__gmpq_set_den(&mut self.0, integer::raw(den)) };
    }

    pub fn num_ref(&self) -> &Integer {
        let num = &(self.0)._mp_num as *const gmp::__mpz_struct as *const Integer;
        unsafe { &*num }
    }

    pub fn den_ref(&self) -> &Integer {
        let den = &(self.0)._mp_den as *const gmp::__mpz_struct as *const Integer;
        unsafe { &*den }
    }

    pub fn num_den_ref(&self) -> (&Integer, &Integer) {
        let num = &(self.0)._mp_num as *const gmp::__mpz_struct as *const Integer;
        let den = &(self.0)._mp_den as *const gmp::__mpz_struct as *const Integer;
        unsafe { (&*num, &*den) }
    }

    pub fn num_mut(&mut self) -> &mut Integer {
        let num = &mut (self.0)._mp_num as *mut gmp::__mpz_struct as *mut Integer;
        unsafe { &mut *num }
    }

    pub fn den_mut(&mut self) -> &mut Integer {
        let den = &mut (self.0)._mp_den as *mut gmp::__mpz_struct as *mut Integer;
        unsafe { &mut *den }
    }

    pub fn num_den_mut(&mut self) -> (&mut Integer, &mut Integer) {
        let num = &mut (self.0)._mp_num as *mut gmp::__mpz_struct as *mut Integer;
        let den = &mut (self.0)._mp_den as *mut gmp::__mpz_struct as *mut Integer;
        unsafe { (&mut *num, &mut *den) }
    }
}

macro_rules! arith_op {
    ($imp:ident $method:ident,
     $imp_assign:ident $method_assign:ident,
     $func:path) => {
        impl<'a> $imp<&'a Rational> for Rational {
            type Output = Rational;
            fn $method(mut self, op: &'a Rational) -> Rational {
                unsafe { $func(&mut self.0, &self.0, &op.0) };
                self
            }
        }

        impl $imp<Rational> for Rational {
            type Output = Rational;
            fn $method(self, op: Rational) -> Rational {
                self.$method(&op)
            }
        }

        impl<'a> $imp_assign<&'a Rational> for Rational {
            fn $method_assign(&mut self, op: &'a Rational) {
                unsafe { $func(&mut self.0, &self.0, &op.0) };
            }
        }

        impl $imp_assign<Rational> for Rational {
            fn $method_assign(&mut self, op: Rational) {
                self.add_assign(&op);
            }
        }
    }
}

arith_op!{ Add add, AddAssign add_assign, gmp::__gmpq_add }
arith_op!{ Sub sub, SubAssign sub_assign, gmp::__gmpq_sub }
arith_op!{ Mul mul, MulAssign mul_assign, gmp::__gmpq_mul }
arith_op!{ Div div, DivAssign div_assign, gmp::__gmpq_div }

impl Shl<u64> for Rational {
    type Output = Rational;
    fn shl(mut self, op: u64) -> Rational {
        unsafe { gmp::__gmpq_mul_2exp(&mut self.0, &self.0, op) };
        self
    }
}

impl ShlAssign<u64> for Rational {
    fn shl_assign(&mut self, op: u64) {
        unsafe { gmp::__gmpq_mul_2exp(&mut self.0, &self.0, op) };
    }
}

impl Shr<u64> for Rational {
    type Output = Rational;
    fn shr(mut self, op: u64) -> Rational {
        unsafe { gmp::__gmpq_div_2exp(&mut self.0, &self.0, op) };
        self
    }
}

impl ShrAssign<u64> for Rational {
    fn shr_assign(&mut self, op: u64) {
        unsafe { gmp::__gmpq_div_2exp(&mut self.0, &self.0, op) };
    }
}

impl Neg for Rational {
    type Output = Rational;
    fn neg(mut self) -> Rational {
        unsafe { gmp::__gmpq_neg(&mut self.0, &self.0) };
        self
    }
}

impl Rational {
    fn abs(mut self) -> Rational {
        unsafe { gmp::__gmpq_abs(&mut self.0, &self.0) };
        self
    }

    fn inv(mut self) -> Rational {
        unsafe { gmp::__gmpq_inv(&mut self.0, &self.0) };
        self
    }
}

impl Eq for Rational {}

impl Ord for Rational {
    fn cmp(&self, other: &Rational) -> Ordering {
        let ord = unsafe { gmp::__gmpq_cmp(&self.0, &other.0) };
        ord.cmp(&0)
    }
}

impl PartialEq for Rational {
    fn eq(&self, other: &Rational) -> bool {
        unsafe { gmp::__gmpq_equal(&self.0, &other.0) != 0 }
    }
}

impl PartialOrd for Rational {
    fn partial_cmp(&self, other: &Rational) -> Option<Ordering> {
        Some(self.cmp(other))
    }
}

impl PartialEq<Integer> for Rational {
    fn eq(&self, other: &Integer) -> bool {
        self.partial_cmp(other) == Some(Ordering::Equal)
    }
}

impl PartialOrd<Integer> for Rational {
    fn partial_cmp(&self, other: &Integer) -> Option<Ordering> {
        let ord = unsafe { gmp::__gmpq_cmp_z(&self.0, integer::raw(other)) };
        Some(ord.cmp(&0))
    }
}

impl PartialEq<Rational> for Integer {
    fn eq(&self, other: &Rational) -> bool {
        other.eq(self)
    }
}

impl PartialOrd<Rational> for Integer {
    fn partial_cmp(&self, other: &Rational) -> Option<Ordering> {
        match other.partial_cmp(self) {
            None => None,
            Some(x) => Some(x.reverse()),
        }
    }
}

impl PartialEq<(u64, u64)> for Rational {
    fn eq(&self, other: &(u64, u64)) -> bool {
        self.partial_cmp(other) == Some(Ordering::Equal)
    }
}

impl PartialOrd<(u64, u64)> for Rational {
    fn partial_cmp(&self, other: &(u64, u64)) -> Option<Ordering> {
        let ord = unsafe { gmp::__gmpq_cmp_ui(&self.0, other.0, other.1) };
        Some(ord.cmp(&0))
    }
}

impl PartialEq<Rational> for (u64, u64) {
    fn eq(&self, other: &Rational) -> bool {
        other.eq(self)
    }
}

impl PartialOrd<Rational> for (u64, u64) {
    fn partial_cmp(&self, other: &Rational) -> Option<Ordering> {
        match other.partial_cmp(self) {
            None => None,
            Some(x) => Some(x.reverse()),
        }
    }
}

impl PartialEq<(i64, u64)> for Rational {
    fn eq(&self, other: &(i64, u64)) -> bool {
        self.partial_cmp(other) == Some(Ordering::Equal)
    }
}

impl PartialOrd<(i64, u64)> for Rational {
    fn partial_cmp(&self, other: &(i64, u64)) -> Option<Ordering> {
        let ord = unsafe { gmp::__gmpq_cmp_si(&self.0, other.0, other.1) };
        Some(ord.cmp(&0))
    }
}

impl PartialEq<Rational> for (i64, u64) {
    fn eq(&self, other: &Rational) -> bool {
        other.eq(self)
    }
}

impl PartialOrd<Rational> for (i64, u64) {
    fn partial_cmp(&self, other: &Rational) -> Option<Ordering> {
        match other.partial_cmp(self) {
            None => None,
            Some(x) => Some(x.reverse()),
        }
    }
}

impl Rational {
    fn sgn(&self) -> Ordering {
        if (self.0)._mp_num._mp_size < 0 {
            Ordering::Less
        } else if (self.0)._mp_num._mp_size > 0 {
            Ordering::Greater
        } else {
            Ordering::Equal
        }
    }
}

impl Rational {
    pub fn to_string(&self, base: c_int) -> String {
        self.make_string(base, false)
    }

    fn make_string(&self, base: c_int, to_upper: bool) -> String {
        let s;
        let cstr;
        unsafe {
            s = gmp::__gmpq_get_str(ptr::null_mut(), base, &self.0);
            assert!(!s.is_null());
            cstr = ffi::CStr::from_ptr(s);
        }
        let mut buf = cstr.to_str().unwrap().to_string();
        unsafe {
            let mut free = None;
            gmp::__gmp_get_memory_functions(ptr::null_mut(), ptr::null_mut(), &mut free);
            let free = free.unwrap();
            let free_len = cstr.to_bytes().len() + 1;
            free(s as *mut c_void, free_len);
        }
        if to_upper {
            buf = buf.to_uppercase();
        }
        buf
    }
}

impl fmt::Display for Rational {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        write!(f, "{}", self.to_string(10))
    }
}

impl fmt::Debug for Rational {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        write!(f, "0x{}", self.make_string(16, false))
    }
}

impl fmt::Binary for Rational {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        let prefix = if f.alternate() { "0b" } else { "" };
        write!(f, "{}{}", prefix, self.make_string(2, false))
    }
}

impl fmt::Octal for Rational {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        let prefix = if f.alternate() { "0o" } else { "" };
        write!(f, "{}{}", prefix, self.make_string(8, false))
    }
}

impl fmt::LowerHex for Rational {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        let prefix = if f.alternate() { "0x" } else { "" };
        write!(f, "{}{}", prefix, self.make_string(16, false))
    }
}

impl fmt::UpperHex for Rational {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        let prefix = if f.alternate() { "0X" } else { "" };
        write!(f, "{}{}", prefix, self.make_string(16, true))
    }
}
