// Copyright (C) 2016  University of Malta

// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

use std::io;
use std::fs;

use std::io::Write;

use super::Overwrite;
use super::error::Error;
use super::package::{Direction, SyntaxNodeKind, Tree};

type Writer = io::BufWriter<fs::File>;

pub fn output_component(tree: &Tree, index: usize, overwrite: Overwrite) -> Result<(), Error> {
    let name: &str = match tree[index].kind {
        SyntaxNodeKind::Component(ref name) => name,
        _ => unreachable!(),
    };
    let filename = format!("{}.vhdl", name);
    let res = create_component_file(&filename, overwrite)
        .and_then(|w| write_component_file(w, tree, index));
    match res {
        Ok(_) => Ok(()),
        Err(e) => Err(Error::io(&filename, e)),
    }
}

fn create_component_file(filename: &str, overwrite: Overwrite) -> Result<Writer, io::Error> {
    Ok(Writer::new(fs::OpenOptions::new().write(true)
        .create(true)
        .truncate(true)
        .create_new(!overwrite.0)
        .open(&filename)?))
}

fn write_component_file(mut w: Writer, tree: &Tree, comp: usize) -> Result<(), io::Error> {
    let name = match tree[comp].kind {
        SyntaxNodeKind::Component(ref name) => name,
        _ => unreachable!(),
    };
    write!(w,
           concat!("-- generated by rox\n",
                   "\n",
                   "library ieee;\n",
                   "use ieee.std_logic_1164.all;\n",
                   "use ieee.numeric_std.all;\n",
                   "\n",
                   "entity {} is\n\n"),
           name)?;
    let par_list_index = tree.child(comp).unwrap();
    write_port_map(&mut w, tree, par_list_index)?;
    write!(w, "end entity {};\n\narchitecture rox of {0} is\n\n", name)?;
    write_arch_signals(&mut w, tree, par_list_index)?;
    write!(w, "begin  -- architecture rox\n\n")?;
    write_arch_io(&mut w, tree, par_list_index)?;
    let body_index = tree.sibling(par_list_index).unwrap();
    write_arch_body(&mut w, tree, body_index)?;
    write!(w, "end architecture rox;\n")?;
    Ok(())
}

fn write_port_map(w: &mut Writer, tree: &Tree, par_list_index: usize) -> Result<(), io::Error> {
    assert!(tree[par_list_index].kind == SyntaxNodeKind::ParameterList);
    let mut first_par = true;
    let mut par_index_o = tree.child(par_list_index);
    while let Some(par_index) = par_index_o {
        let par_name = match tree[par_index].kind {
            SyntaxNodeKind::Parameter(ref name) => name,
            _ => unreachable!(),
        };
        let dir_index = tree.child(par_index).unwrap();
        let direction = match tree[dir_index].kind {
            SyntaxNodeKind::Direction(Direction::In) => "in",
            SyntaxNodeKind::Direction(Direction::Out) => "out",
            SyntaxNodeKind::Direction(Direction::Bidir) => "inout",
            _ => unreachable!(),
        };
        let type_index = tree.sibling(dir_index).unwrap();
        assert!(tree[type_index].kind == SyntaxNodeKind::ExpressionParameters);
        let ident_index = tree.child(type_index).unwrap();
        match tree[ident_index].kind {
            SyntaxNodeKind::Identifier(ref type_name) => {
                assert!(type_name == "unsigned" || type_name == "signed")
            }
            _ => unreachable!(),
        };
        let type_arg_index = tree.sibling(ident_index).unwrap();
        let size = match tree[type_arg_index].kind {
            SyntaxNodeKind::IntegerLiteral(ref value) => value,
            _ => unreachable!(),
        };
        if first_par {
            write!(w, "    port(")?;
        } else {
            write!(w, ";\n         ")?;
        }
        write!(w, "{}_: {} std_logic_vector(", par_name, direction)?;
        if *size == 0u64 {
            write!(w, "-1 downto 0)")?;
        } else {
            write!(w, "{} downto 0)", size.clone() - 1)?;
        }
        first_par = false;
        par_index_o = tree.sibling(par_index);
    }
    if !first_par {
        write!(w, ");\n\n")?;
    }
    Ok(())
}

fn write_arch_signals(w: &mut Writer, tree: &Tree, par_list_index: usize) -> Result<(), io::Error> {
    assert!(tree[par_list_index].kind == SyntaxNodeKind::ParameterList);
    let mut first_par = true;
    let mut par_index_o = tree.child(par_list_index);
    while let Some(par_index) = par_index_o {
        let par_name = match tree[par_index].kind {
            SyntaxNodeKind::Parameter(ref name) => name,
            _ => unreachable!(),
        };
        let dir_index = tree.child(par_index).unwrap();
        let direction = match tree[dir_index].kind {
            SyntaxNodeKind::Direction(d) => d,
            _ => unreachable!(),
        };
        let type_index = tree.sibling(dir_index).unwrap();
        assert!(tree[type_index].kind == SyntaxNodeKind::ExpressionParameters);
        let ident_index = tree.child(type_index).unwrap();
        let signed = match tree[ident_index].kind {
            SyntaxNodeKind::Identifier(ref type_name) => {
                if type_name == "unsigned" {
                    false
                } else if type_name == "signed" {
                    true
                } else {
                    unreachable!()
                }
            }
            _ => unreachable!(),
        };
        let type_arg_index = tree.sibling(ident_index).unwrap();
        let size = match tree[type_arg_index].kind {
            SyntaxNodeKind::IntegerLiteral(ref value) => value,
            _ => unreachable!(),
        };
        if direction == Direction::In || direction == Direction::Bidir {
            write!(w,
                   "    signal i_{}: {}(",
                   par_name,
                   if signed { "signed" } else { "unsigned" })?;
            if *size == 0u64 {
                write!(w, "-1 downto 0);\n")?;
            } else {
                write!(w, "{} downto 0);\n", size.clone() - 1)?;
            }
        }
        if direction == Direction::Out || direction == Direction::Bidir {
            write!(w,
                   "    signal o_{}: {}(",
                   par_name,
                   if signed { "signed" } else { "unsigned" })?;
            if *size == 0u64 {
                write!(w, "-1 downto 0);\n")?;
            } else {
                write!(w, "{} downto 0);\n", size.clone() - 1)?;
            }
        }
        first_par = false;
        par_index_o = tree.sibling(par_index);
    }
    if !first_par {
        write!(w, "\n")?;
    }
    Ok(())
}

fn write_arch_io(w: &mut Writer, tree: &Tree, par_list_index: usize) -> Result<(), io::Error> {
    assert!(tree[par_list_index].kind == SyntaxNodeKind::ParameterList);
    let mut first_par = true;
    let mut par_index_o = tree.child(par_list_index);
    while let Some(par_index) = par_index_o {
        let par_name = match tree[par_index].kind {
            SyntaxNodeKind::Parameter(ref name) => name,
            _ => unreachable!(),
        };
        let dir_index = tree.child(par_index).unwrap();
        let direction = match tree[dir_index].kind {
            SyntaxNodeKind::Direction(d) => d,
            _ => unreachable!(),
        };
        let type_index = tree.sibling(dir_index).unwrap();
        assert!(tree[type_index].kind == SyntaxNodeKind::ExpressionParameters);
        let ident_index = tree.child(type_index).unwrap();
        let signed = match tree[ident_index].kind {
            SyntaxNodeKind::Identifier(ref type_name) => {
                if type_name == "unsigned" {
                    false
                } else if type_name == "signed" {
                    true
                } else {
                    unreachable!()
                }
            }
            _ => unreachable!(),
        };
        if direction == Direction::In || direction == Direction::Bidir {
            write!(w,
                   "    i_{} <= {}({0}_);\n",
                   par_name,
                   if signed { "signed" } else { "unsigned" })?;
        }
        if direction == Direction::Out || direction == Direction::Bidir {
            write!(w, "    {}_ <= std_logic_vector(o_{0});\n", par_name)?;
        }
        first_par = false;
        par_index_o = tree.sibling(par_index);
    }
    if !first_par {
        write!(w, "\n")?;
    }
    Ok(())
}

fn write_arch_body(w: &mut Writer, tree: &Tree, body_index: usize) -> Result<(), io::Error> {
    assert!(tree[body_index].kind == SyntaxNodeKind::Body);
    let mut first_stat = true;
    let mut stat_index_o = tree.child(body_index);
    while let Some(stat_index) = stat_index_o {
        let a = match tree[stat_index].kind {
            SyntaxNodeKind::Assignment(ref a) => a,
            _ => unreachable!(),
        };
        write!(w, "    -- {} = ...\n", a)?;
        first_stat = false;
        stat_index_o = tree.sibling(stat_index);
    }
    if !first_stat {
        write!(w, "\n")?;
    }
    Ok(())
}
