// Copyright (C) 2016  University of Malta

// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

use std::fmt;

use super::output;
use super::parse;
use super::pass;
use super::tree;

use super::Overwrite;
use super::error::Error;
use super::input::{SourceSet, Span};
use super::number::{Float, Integer};
use super::parse::TokenKind;
use super::tree::Insertion;

#[derive(Clone, PartialEq)]
pub struct Package {
    sources: SourceSet,
    tree: Tree,
}

pub type Tree = tree::Tree<SyntaxNode>;

#[derive(Clone, PartialEq, Debug)]
pub struct SyntaxNode {
    pub kind: SyntaxNodeKind,
    pub span: Option<Span>,
}

#[derive(Clone, PartialEq, Debug)]
pub enum SyntaxNodeKind {
    Package,
    Component(String),
    Constant(String),
    IntegerLiteral(Integer),
    IntegerXzLiteral(Integer, Integer), // [i] = if .1[i]==0 {.0[i]} else if .0[i]==0 {x} else {z}
    FloatLiteral(Float),
    DontCare,
    HighImpedance,
    Identifier(String),
    ParameterList,
    Parameter(String),
    Direction(Direction),
    Body,
    Assignment(String),
    ExpressionRelational(TokenKind),
    ExpressionBitwise(TokenKind),
    ExpressionShift(TokenKind),
    ExpressionAddition(TokenKind),
    ExpressionMultiplication(TokenKind),
    ExpressionExponentiation(TokenKind),
    ExpressionCast,
    ExpressionUnary(TokenKind),
    ExpressionGroup,
    ExpressionField,
    ExpressionParameters,
    ExpressionIndex,
    Empty,
}

#[derive(Copy, Clone, PartialEq, Debug)]
pub enum Direction {
    In,
    Out,
    Bidir,
}

impl Package {
    pub fn new() -> Package {
        Package {
            sources: SourceSet::new(),
            tree: Tree::new(),
        }
    }

    pub fn read_source(&mut self, filename: &str) -> Result<(), Error> {
        self.sources.read(filename)
    }

    pub fn parse(&mut self) -> Result<(), Error> {
        let node = SyntaxNode {
            kind: SyntaxNodeKind::Package,
            span: None,
        };
        self.tree.insert(Insertion::Root, node);
        for mut span in self.sources.source_spans() {
            let mut insertion = Insertion::ChildOf(self.tree.root().unwrap());
            loop {
                span.skip_whitespace(&self.sources);
                if span.start == span.end {
                    break;
                }
                let comp = parse::parse(&mut self.tree, &self.sources, &mut span, insertion)?;
                insertion = Insertion::SiblingOf(comp);
            }
        }
        Ok(())
    }

    pub fn process(&mut self) -> Result<(), Error> {
        pass::evaluate_constants(&mut self.tree);
        Ok(())
    }

    pub fn output(&self, overwrite: Overwrite) -> Result<(), Error> {
        let root = self.tree.root().unwrap();
        let mut ocomp = self.tree.child(root);
        while let Some(comp) = ocomp {
            if let SyntaxNodeKind::Component(_) = self.tree[comp].kind {
                output::output_component(&self.tree, comp, overwrite)?;
            }
            ocomp = self.tree.sibling(comp);
        }
        Ok(())
    }
}

impl fmt::Debug for Package {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        write!(f, "Sources:")?;
        for s in self.sources.source_spans() {
            write!(f, " {}", s.filename(&self.sources))?;
        }
        if let Some(root) = self.tree.root() {
            format_node(&self.tree, &self.sources, f, root, "")
        } else {
            write!(f, "empty")
        }
    }
}

fn format_node(tree: &Tree,
               sources: &SourceSet,
               f: &mut fmt::Formatter,
               index: usize,
               prefix: &str)
               -> fmt::Result {
    let mut child = tree.child(index);
    let sibling = tree.sibling(index);
    let second_prefix = if index == 0 {
        ""
    } else if sibling.is_some() {
        "├── "
    } else {
        "└── "
    };
    write!(f, "\n{}{}{:?}", prefix, second_prefix, tree[index].kind)?;
    if let Some(span) = tree[index].span {
        write!(f, " \"{}\"", &squash_string(span.get(sources)))?;
    }
    let second_prefix = if index == 0 {
        ""
    } else if sibling.is_some() {
        "│   "
    } else {
        "    "
    };
    let mut child_prefix = prefix.to_string();
    child_prefix.push_str(second_prefix);
    while let Some(c) = child {
        format_node(tree, sources, f, c, &child_prefix)?;
        child = tree.sibling(c);
    }
    Ok(())
}

fn squash_string(s: &str) -> String {
    let char_count = s.chars().count();
    let s: String = if char_count > 32 {
        let mut buf = String::new();
        let mut iter = s.chars();
        for _ in 0..24 {
            buf.push(iter.next().unwrap());
        }
        for _ in 24..char_count - 5 {
            iter.next();
        }
        buf.push_str("...");
        for c in iter {
            buf.push(c);
        }
        buf
    } else {
        s.into()
    };
    s.replace('\n', "\\n")
}
