// Copyright (C) 2016  University of Malta

// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

use std::cmp;

use super::number;

use super::error::Error;
use super::input::{SourceSet, Span};
use super::number::{Float, FromPrec, Integer, Pow};
use super::package::{Direction, SyntaxNode, SyntaxNodeKind, Tree};
use super::tree::Insertion;

#[derive(Copy, Clone, PartialEq, Debug)]
pub enum TokenKind {
    Arrow, // =>
    Equals, // =
    Comma, // ,
    Semicolon, // ;
    Colon, // :

    // increasing in precedence
    IsEquals, // ==
    NotEquals, // !=
    Less, // <
    LessEquals, // <=
    Greater, // >
    GreaterEquals, // >=

    And, // &
    Or, // |
    Xor, // ^
    Nand, // !&
    Nor, // !|
    Xnor, // !^

    ShiftLeft, // <<
    ShiftRight, // >>

    Plus, // +
    Minus, // -

    Times, // *
    Division, // /
    Remainder, // %

    Power, // **

    // Other unary operators at same level of precedence: & | ^ + -
    // Note that !& !| !^ do not need unary operators as for example
    // !(&x) would be equal to !&x.
    Not, // !

    Point, // .
    OpenParens, // (
    CloseParens, // )
    OpenSquare, // [
    CloseSquare, // ]
    OpenBrace, // {
    CloseBrace, // }
    Integer,
    IntegerXz,
    Float,
    ZeroX,
    ZeroZ,
    Word,
    Eof,
    Unknown,
}

#[derive(Copy, Clone, PartialEq, Debug)]
struct Token {
    kind: TokenKind,
    span: Span,
}

// first character is known to be _ | a...z | A...Z
fn get_word_chars_bytes(s: &str) -> (TokenKind, usize, usize) {
    let mut iter = s.char_indices();

    let (mut chars, mut bytes) = match iter.next() {
        Some((index, c)) if c == '_' || ('a' <= c && c <= 'z') || ('A' <= c && c <= 'Z') => {
            (1, index + c.len_utf8())
        }
        _ => unreachable!(),
    };
    for (index, c) in iter {
        if c == '_' || ('a' <= c && c <= 'z') || ('A' <= c && c <= 'Z') || ('0' <= c && c <= '9') {
            chars += 1;
            bytes = index + c.len_utf8();
        } else {
            break;
        }
    }
    (TokenKind::Word, chars, bytes)
}

// Point, ZeroZ, ZeroX, Integer or Float
fn get_number_chars_bytes(s: &str) -> Result<(TokenKind, usize, usize), (String, usize, usize)> {
    let mut base = 10;
    let mut has_point = false;
    let mut iter = s.char_indices();
    let (mut prev, mut chars, mut bytes) = match iter.next() {
        Some((index, c)) if c == '.' => {
            has_point = true;
            match iter.next() {
                Some((index, c)) if '0' <= c && c <= '9' => (c, 2, index + c.len_utf8()),
                _ => return Ok((TokenKind::Point, 1, index + c.len_utf8())),
            }
        }
        Some((index0, c0)) if c0 == '0' => {
            if let Some((index1, c1)) = iter.next() {
                match c1 {
                    '0'...'9' | '_' => {}
                    '.' => has_point = true,
                    'b' => base = 2,
                    'q' => base = 4,
                    'o' => base = 8,
                    'x' => base = 16,
                    'z' => return Ok((TokenKind::ZeroZ, 2, index1 + c1.len_utf8())),
                    _ => return Ok((TokenKind::Integer, 1, index0 + c0.len_utf8())),
                }
                (c1, 2, index1 + c1.len_utf8())
            } else {
                return Ok((TokenKind::Integer, 1, index0 + c0.len_utf8()));
            }
        }
        Some((index, c)) if '1' <= c && c <= '9' => (c, 1, index + c.len_utf8()),
        _ => unreachable!(),
    };
    let mut has_digits = base == 10;
    let mut has_mode = false;
    let mut has_exponent = false;
    while let Some((index, c)) = iter.next() {
        let (prev_chars, prev_bytes) = (chars, bytes);
        chars += 1;
        bytes = index + c.len_utf8();
        match c {
            '_' if prev != '.' => {}
            '0'...'9' if c as i32 - '0' as i32 >= base => {
                return Err((format!("digit '{}' too large for base {}", c, base), chars, bytes))
            }
            '0'...'9' => has_digits = true,
            'a'...'f' | 'A'...'F' if base == 16 => has_digits = true,
            'x' | 'z' if base != 10 && !has_point => {
                has_digits = true;
                has_mode = true;
            }
            '.' if prev != '_' && !has_mode && !has_point => {
                has_point = true;
            }
            'e' | 'E' if base == 10 => {
                has_exponent = true;
                break;
            }
            'p' | 'P' if base != 10 && !has_mode => {
                if !has_digits {
                    return Err(("cannot have floating-point literal with no digits".into(),
                                chars,
                                bytes));
                }
                has_exponent = true;
                break;
            }
            _ => {
                chars = prev_chars;
                bytes = prev_bytes;
                break;
            }
        }
        prev = c;
    }
    if has_exponent {
        let (index, c) = match iter.next() {
            Some((index, c)) if c == '+' || c == '-' => (index, c),
            Some((index, c)) if '0' <= c && c <= '9' => (index, c),
            _ => return Err(("exponent must have at least one decimal digit".into(), chars, bytes)),
        };
        has_digits = '0' <= c && c <= '9';
        chars += 1;
        bytes = index + c.len_utf8();
        for (index, c) in iter {
            if '0' <= c && c <= '9' {
                chars += 1;
                bytes = index + c.len_utf8();
                has_digits = true;
            } else if !has_digits {
                return Err(("exponent must have  at least one decimal digit".into(), chars, index));
            } else if c == '_' {
                chars += 1;
                bytes = index + c.len_utf8();
            } else {
                break;
            }
        }
    }
    if has_exponent || has_point {
        Ok((TokenKind::Float, chars, bytes))
    } else if has_mode {
        Ok((TokenKind::IntegerXz, chars, bytes))
    } else if has_digits {
        Ok((TokenKind::Integer, chars, bytes))
    } else if base == 16 && chars == 2 {
        // 0x is valid without extra digits, but it cannot have trailing underscores
        Ok((TokenKind::ZeroX, chars, bytes))
    } else {
        Err(("number literal must have at least one digit".into(), chars, bytes))
    }
}

impl Token {
    fn get<'a>(&self, sources: &'a SourceSet) -> &'a str {
        self.span.get(sources)
    }

    fn err(&self, sources: &SourceSet, expected: &str) -> Error {
        let found_str = format!("'{}'", self.get(sources));
        let found = if self.kind == TokenKind::Eof {
            "end of file"
        } else {
            &found_str
        };
        Error::parse(&format!("{}:{}:{} Found {}, expected {}",
                              self.span.filename(sources),
                              self.span.line,
                              self.span.column,
                              found,
                              expected))
    }

    fn err_message(&self, sources: &SourceSet, message: &str) -> Error {
        let found_str = format!("'{}'", self.get(sources));
        let found = if self.kind == TokenKind::Eof {
            "end of file"
        } else {
            &found_str
        };
        Error::parse(&format!("{}:{}:{} Found {}, {}",
                              self.span.filename(sources),
                              self.span.line,
                              self.span.column,
                              found,
                              message))
    }

    fn next(span: &mut Span, sources: &SourceSet) -> Result<Token, Error> {
        let stream = span.get_whole(sources);
        span.skip_whitespace(sources);
        let mut ret = Token {
            kind: TokenKind::Unknown,
            span: *span,
        };
        let mut iter = stream[span.start..span.end].char_indices();
        let (first_c, first_end) = match iter.next() {
            Some((index, c)) => (c, span.start + index + c.len_utf8()),
            None => {
                ret.kind = TokenKind::Eof;
                return Ok(ret);
            }
        };
        span.column += 1;
        ret.span.end = first_end;

        // check for two-character symbols
        if let Some((index, second_c)) = iter.next() {
            ret.kind = match (first_c, second_c) {
                ('=', '>') => TokenKind::Arrow,
                ('=', '=') => TokenKind::IsEquals,
                ('!', '=') => TokenKind::NotEquals,
                ('<', '=') => TokenKind::LessEquals,
                ('>', '=') => TokenKind::GreaterEquals,
                ('!', '&') => TokenKind::Nand,
                ('!', '|') => TokenKind::Nor,
                ('!', '^') => TokenKind::Xnor,
                ('<', '<') => TokenKind::ShiftLeft,
                ('>', '>') => TokenKind::ShiftRight,
                ('*', '*') => TokenKind::Power,
                _ => TokenKind::Unknown,
            };
            if ret.kind != TokenKind::Unknown {
                span.start = span.start + index + second_c.len_utf8();
                span.column += 1;
                ret.span.end = span.start;
                return Ok(ret);
            }
        }

        ret.kind = match first_c {
            '=' => TokenKind::Equals,
            ',' => TokenKind::Comma,
            ';' => TokenKind::Semicolon,
            ':' => TokenKind::Colon,
            '<' => TokenKind::Less,
            '>' => TokenKind::Greater,
            '&' => TokenKind::And,
            '|' => TokenKind::Or,
            '^' => TokenKind::Xor,
            '+' => TokenKind::Plus,
            '-' => TokenKind::Minus,
            '*' => TokenKind::Times,
            '/' => TokenKind::Division,
            '%' => TokenKind::Remainder,
            '!' => TokenKind::Not,
            '(' => TokenKind::OpenParens,
            ')' => TokenKind::CloseParens,
            '[' => TokenKind::OpenSquare,
            ']' => TokenKind::CloseSquare,
            '{' => TokenKind::OpenBrace,
            '}' => TokenKind::CloseBrace,
            '0'...'9' | '.' => {
                let (kind, chars, bytes) =
                    match get_number_chars_bytes(&stream[span.start..span.end]) {
                        Ok(x) => x,
                        Err((msg, _, bytes)) => {
                            let mut err_span = *span;
                            err_span.end = span.start + bytes;
                            return Err(err_span.err(sources, &msg));
                        }
                    };
                span.column += chars - 1;
                ret.span.end = span.start + bytes;
                kind
            }
            '_' | 'a'...'z' | 'A'...'Z' => {
                let (kind, chars, bytes) = get_word_chars_bytes(&stream[span.start..span.end]);
                span.column += chars - 1;
                ret.span.end = span.start + bytes;
                kind
            }
            _ => TokenKind::Unknown,
        };
        span.start = ret.span.end;
        Ok(ret)
    }

    // this must not be called on anything except TokenKind::Integer and TokenKind::IntegerXz
    fn integer_literal(&self, sources: &SourceSet) -> SyntaxNodeKind {
        let mut base = 10;
        let mut iter = self.get(sources).chars();
        let mut value = iter.next().unwrap() as u64 - '0' as u64;
        if value == 0 {
            if let Some(c) = iter.next() {
                match c {
                    '0'...'9' => value = c as u64 - '0' as u64,
                    '_' => {}
                    'b' => base = 2,
                    'q' => base = 4,
                    'o' => base = 8,
                    'x' => base = 16,
                    _ => unreachable!(),
                };
            } else {
                return SyntaxNodeKind::IntegerLiteral(Integer::new());
            }
        }
        let mut mode = 0u64;
        let mut ret = (Integer::new(), Integer::new());
        let mut counter = 1_u64;
        for c in iter {
            let (cvalue, cmode) = match c {
                '_' => continue,
                '0'...'9' => (c as u64 - '0' as u64, 0),
                'a'...'f' => (c as u64 - 'a' as u64 + 10, 0),
                'A'...'F' => (c as u64 - 'A' as u64 + 10, 0),
                'x' => (0, base - 1),
                'z' => (base - 1, base - 1),
                _ => unreachable!(),
            };
            value = value * base + cvalue;
            mode = mode * base + cmode;
            counter += 1;
            match (base, counter) {
                (2, 64) | (4, 32) | (16, 16) => {
                    ret = ((ret.0 << 64) + value, (ret.1 << 64) + mode);
                    value = 0;
                    mode = 0;
                    counter = 0;
                }
                (8, 21) => {
                    ret = ((ret.0 << 63) + value, (ret.1 << 63) + mode);
                    value = 0;
                    mode = 0;
                    counter = 0;
                }
                (10, 19) => {
                    ret.0 = ret.0 * 10_000_000_000_000_000_000_u64 + value;
                    value = 0;
                    counter = 0;
                }
                _ => {}
            };
        }
        ret = match base {
            2 => ((ret.0 << counter) + value, (ret.1 << counter) + mode),
            4 => ((ret.0 << (counter / 2)) + value, (ret.1 << (counter / 2)) + mode),
            8 => ((ret.0 << (counter / 3)) + value, (ret.1 << (counter / 3)) + mode),
            10 => (ret.0 * 10_u64.pow(counter as u32) + value, ret.1),
            16 => ((ret.0 << (counter / 4)) + value, (ret.1 << (counter / 4)) + mode),
            _ => unreachable!(),
        };
        if ret.1 != 0u64 {
            SyntaxNodeKind::IntegerXzLiteral(ret.0, ret.1)
        } else {
            SyntaxNodeKind::IntegerLiteral(ret.0)
        }
    }

    // this must not be called on anything except TokenKind::Float
    fn float_literal(&self, sources: &SourceSet) -> SyntaxNodeKind {
        let mut base = 10;
        let mut has_point = false;
        let mut value = 0;
        let mut digits_after_point = 0;
        let mut iter = self.get(sources).chars();
        match iter.next().unwrap() {
            '.' => has_point = true,
            '0' => {
                match iter.next() {
                    Some(c) => {
                        match c {
                            '_' => {}
                            '0'...'9' => value = c as u64 - '0' as u64,
                            '.' => has_point = true,
                            'b' => base = 2,
                            'q' => base = 4,
                            'o' => base = 8,
                            'x' => base = 16,
                            _ => unreachable!(),
                        }
                    }
                    None => return SyntaxNodeKind::FloatLiteral(Float::new(number::Prec(53))),
                }
            }
            c @ '1'...'9' => value = c as u64 - '0' as u64,
            _ => unreachable!(),
        }
        let mut significant_digits = if value == 0 { 0 } else { 1 };
        let mut significand = Integer::new();
        let mut counter = 1_u64;
        while let Some(c) = iter.next() {
            let cvalue = match c {
                '_' => continue,
                '.' => {
                    has_point = true;
                    continue;
                }
                '0'...'9' => c as u64 - '0' as u64,
                'a'...'f' if base == 16 => c as u64 - 'a' as u64 + 10,
                'A'...'F' if base == 16 => c as u64 - 'A' as u64 + 10,
                'e' | 'E' | 'p' | 'P' => break,
                _ => unreachable!(),
            };
            if cvalue != 0 || significant_digits != 0 {
                significant_digits += 1;
            }
            if has_point {
                digits_after_point += 1;
            }
            value = value * base + cvalue;
            counter += 1;
            match (base, counter) {
                (2, 64) | (4, 32) | (16, 16) => {
                    significand = (significand << 64) + value;
                    value = 0;
                    counter = 0;
                }
                (8, 21) => {
                    significand = (significand << 63) + value;
                    value = 0;
                    counter = 0;
                }
                (10, 19) => {
                    significand = significand * 10_000_000_000_000_000_000_u64 + value;
                    value = 0;
                    counter = 0;
                }
                _ => {}
            };
        }
        significand = match base {
            2 => (significand << counter) + value,
            4 => (significand << (counter / 2)) + value,
            8 => (significand << (counter / 3)) + value,
            10 => significand * 10_u64.pow(counter as u32) + value,
            16 => (significand << (counter / 4)) + value,
            _ => unreachable!(),
        };
        let mut exp_sign = false;
        let mut exp = Some(0i64);
        for c in iter {
            match c {
                '+' | '_' => {}
                '-' => exp_sign = true,
                '0'...'9' => {
                    let digit = c as i64 - '0' as i64;
                    exp = exp.and_then(|x| x.checked_mul(10)).and_then(|x| x.checked_add(digit));
                }
                _ => unreachable!(),
            }
        }
        if exp_sign {
            exp = exp.and_then(i64::checked_neg);
        }
        let mut precision = match base {
            2 => significant_digits + 1,
            4 => significant_digits * 2 + 1,
            8 => significant_digits * 3 + 1,
            10 => (significant_digits * 10 + 2) / 3 + 1,
            16 => significant_digits * 4 + 1,
            _ => unreachable!(),
        };
        precision = cmp::max(precision, 64);
        let mut ret = Float::from_prec(number::Prec(precision), significand);
        let exp_per_digit = match base {
            2 | 10 => 1,
            4 => 2,
            8 => 3,
            16 => 4,
            _ => unreachable!(),
        };
        exp = exp.and_then(|x| x.checked_sub(digits_after_point * exp_per_digit));
        match exp {
            Some(exp) if base != 10 => ret <<= exp,
            Some(exp) => ret *= Float::from_prec(number::Prec(precision), 10u64).pow(exp),
            None if exp_sign => ret.set_special(number::Special::Zero),
            None => ret.set_special(number::Special::Infinity),
        }
        SyntaxNodeKind::FloatLiteral(ret)
    }
}

pub fn parse(tree: &mut Tree,
             sources: &SourceSet,
             span: &mut Span,
             insertion: Insertion)
             -> Result<usize, Error> {
    let expected = "'comp' or 'const'";
    let mut peek = *span;
    let token = Token::next(&mut peek, sources)?;
    if token.kind != TokenKind::Word {
        return Err(token.err(sources, expected));
    }
    match token.get(sources) {
        "comp" => parse_comp(tree, sources, span, insertion),
        "const" => parse_const(tree, sources, span, insertion),
        _ => Err(token.err(sources, expected)),
    }

}

fn parse_comp(tree: &mut Tree,
              sources: &SourceSet,
              span: &mut Span,
              insertion: Insertion)
              -> Result<usize, Error> {
    let comp_token = Token::next(span, sources)?;
    if comp_token.kind != TokenKind::Word || comp_token.get(sources) != "comp" {
        unreachable!();
    }
    let token = Token::next(span, sources)?;
    if token.kind != TokenKind::Word {
        return Err(token.err(sources, "component identifier"));
    }
    let comp_node = SyntaxNode {
        kind: SyntaxNodeKind::Component(token.get(sources).to_string()),
        span: Some(comp_token.span),
    };
    let comp_i = tree.insert(insertion, comp_node);
    let params_i = parse_comp_params(tree, sources, span, Insertion::ChildOf(comp_i))?;
    let body_i = parse_comp_body(tree, sources, span, Insertion::SiblingOf(params_i))?;
    tree[comp_i].span.as_mut().unwrap().end = tree[body_i].span.unwrap().end;
    Ok(comp_i)
}

fn parse_comp_params(tree: &mut Tree,
                     sources: &SourceSet,
                     span: &mut Span,
                     mut insertion: Insertion)
                     -> Result<usize, Error> {
    let open = parse_token(sources, span, TokenKind::OpenParens, "'('")?;
    let params_node = SyntaxNode {
        kind: SyntaxNodeKind::ParameterList,
        span: Some(open.span),
    };
    let params_i = tree.insert(insertion, params_node);
    insertion = Insertion::ChildOf(params_i);
    let mut need_comma = false;
    loop {
        let mut token = Token::next(span, sources)?;
        if need_comma {
            match token.kind {
                TokenKind::Comma => token = Token::next(span, sources)?,
                TokenKind::CloseParens => {}
                _ => return Err(token.err(sources, "')' or ','")),
            }
        }
        let ident = match token.kind {
            TokenKind::Word => token.get(sources).to_string(),
            TokenKind::CloseParens => {
                tree[params_i].span.as_mut().unwrap().end = token.span.end;
                return Ok(params_i);
            }
            _ => return Err(token.err(sources, "')' or parameter")),
        };
        let param_node = SyntaxNode {
            kind: SyntaxNodeKind::Parameter(ident),
            span: Some(token.span),
        };
        let param_i = tree.insert(insertion, param_node);
        insertion = Insertion::SiblingOf(param_i);
        parse_token(sources, span, TokenKind::Colon, "':'")?;
        let dir_i = parse_dir(tree, sources, span, Insertion::ChildOf(param_i))?;
        let type_i = parse_expr(tree, sources, span, Insertion::SiblingOf(dir_i))?;
        tree[param_i].span.as_mut().unwrap().end = tree[type_i].span.unwrap().end;
        need_comma = true;
    }
}

fn parse_comp_body(tree: &mut Tree,
                   sources: &SourceSet,
                   span: &mut Span,
                   mut insertion: Insertion)
                   -> Result<usize, Error> {
    let open = parse_token(sources, span, TokenKind::OpenBrace, "'{'")?;
    let body_node = SyntaxNode {
        kind: SyntaxNodeKind::Body,
        span: Some(open.span),
    };
    let body_i = tree.insert(insertion, body_node);
    insertion = Insertion::ChildOf(body_i);
    loop {
        let mut peek = *span;
        if let Ok(close) = parse_token(sources, &mut peek, TokenKind::CloseBrace, "") {
            *span = peek;
            tree[body_i].span.as_mut().unwrap().end = close.span.end;
            return Ok(body_i);
        }
        let stat_i = parse_stat(tree, sources, span, insertion)?;
        insertion = Insertion::SiblingOf(stat_i);
    }
}

fn parse_const(tree: &mut Tree,
               sources: &SourceSet,
               span: &mut Span,
               insertion: Insertion)
               -> Result<usize, Error> {
    let const_token = Token::next(span, sources)?;
    if const_token.kind != TokenKind::Word || const_token.get(sources) != "const" {
        unreachable!();
    }
    let mut token = Token::next(span, sources)?;
    if token.kind != TokenKind::Word {
        return Err(token.err(sources, "constant identifier"));
    }
    let const_node = SyntaxNode {
        kind: SyntaxNodeKind::Constant(token.get(sources).to_string()),
        span: Some(const_token.span),
    };
    let const_i = tree.insert(insertion, const_node);
    token = Token::next(span, sources)?;
    let type_i = match token.kind {
        TokenKind::Colon => {
            let type_i = parse_expr(tree, sources, span, Insertion::ChildOf(const_i))?;
            parse_token(sources, span, TokenKind::Equals, "'='")?;
            type_i
        }
        TokenKind::Equals => {
            let type_node = SyntaxNode {
                kind: SyntaxNodeKind::Empty,
                span: None,
            };
            tree.insert(Insertion::ChildOf(const_i), type_node)
        }
        _ => return Err(token.err(sources, "':' or '='")),
    };
    let value_i = parse_expr(tree, sources, span, Insertion::SiblingOf(type_i))?;
    tree[const_i].span.as_mut().unwrap().end = tree[value_i].span.unwrap().end;
    parse_token(sources, span, TokenKind::Semicolon, "';'")?;
    Ok(const_i)
}

fn parse_stat(tree: &mut Tree,
              sources: &SourceSet,
              span: &mut Span,
              insertion: Insertion)
              -> Result<usize, Error> {
    let token = Token::next(span, sources)?;
    let ident = match token.kind {
        TokenKind::Word => token.get(sources).to_string(),
        _ => return Err(token.err(sources, "identifier")),
    };
    let ass_node = SyntaxNode {
        kind: SyntaxNodeKind::Assignment(ident),
        span: Some(token.span),
    };
    let ass_i = tree.insert(insertion, ass_node);
    parse_token(sources, span, TokenKind::Equals, "'='")?;
    parse_expr(tree, sources, span, Insertion::ChildOf(ass_i))?;
    let semicolon = parse_token(sources, span, TokenKind::Semicolon, "';'")?;
    tree[ass_i].span.as_mut().unwrap().end = semicolon.span.end;
    Ok(ass_i)
}

fn parse_expr(tree: &mut Tree,
              sources: &SourceSet,
              span: &mut Span,
              insertion: Insertion)
              -> Result<usize, Error> {
    parse_expr_relational(tree, sources, span, insertion)
}

fn parse_expr_relational(tree: &mut Tree,
                         sources: &SourceSet,
                         span: &mut Span,
                         insertion: Insertion)
                         -> Result<usize, Error> {
    let left_side_i = parse_expr_bitwise(tree, sources, span, insertion)?;
    let mut peek = *span;
    let token = Token::next(&mut peek, sources)?;
    match token.kind {
        TokenKind::IsEquals | TokenKind::NotEquals | TokenKind::Less | TokenKind::LessEquals |
        TokenKind::Greater | TokenKind::GreaterEquals => *span = peek,
        _ => return Ok(left_side_i),
    }
    let rel_node = SyntaxNode {
        kind: SyntaxNodeKind::ExpressionRelational(token.kind),
        span: tree[left_side_i].span,
    };
    let rel_i = tree.insert(Insertion::ParentOf(left_side_i), rel_node);
    // left_side_i is no longer correct, as it has been displaced
    let right_side_i = parse_expr_bitwise(tree, sources, span, Insertion::ChildOf(rel_i))?;
    tree[rel_i].span.as_mut().unwrap().end = tree[right_side_i].span.unwrap().end;
    Ok(rel_i)
}

fn parse_expr_bitwise(tree: &mut Tree,
                      sources: &SourceSet,
                      span: &mut Span,
                      insertion: Insertion)
                      -> Result<usize, Error> {
    let mut first: Option<TokenKind> = None;
    let mut left_side_i = parse_expr_shift(tree, sources, span, insertion)?;
    loop {
        let mut peek = *span;
        let token = Token::next(&mut peek, sources)?;
        match token.kind {
            TokenKind::And | TokenKind::Or | TokenKind::Xor | TokenKind::Nand |
            TokenKind::Nor | TokenKind::Xnor => *span = peek,
            _ => return Ok(left_side_i),
        }
        if first.is_none() {
            first = Some(token.kind);
        } else if first.unwrap() != token.kind {
            return Err(token.err_message(sources,
                                         "cannot mix different bitwise operators, consider \
                                          using brackets"));
        }
        let bitwise_node = SyntaxNode {
            kind: SyntaxNodeKind::ExpressionBitwise(token.kind),
            span: tree[left_side_i].span,
        };
        let bitwise_i = tree.insert(Insertion::ParentOf(left_side_i), bitwise_node);
        // left_side_i is no longer correct, as it has been displaced
        let right_side_i = parse_expr_shift(tree, sources, span, Insertion::ChildOf(bitwise_i))?;
        tree[bitwise_i].span.as_mut().unwrap().end = tree[right_side_i].span.unwrap().end;
        left_side_i = bitwise_i;
    }
}

fn parse_expr_shift(tree: &mut Tree,
                    sources: &SourceSet,
                    span: &mut Span,
                    insertion: Insertion)
                    -> Result<usize, Error> {
    let mut left_side_i = parse_expr_add(tree, sources, span, insertion)?;
    loop {
        let mut peek = *span;
        let token = Token::next(&mut peek, sources)?;
        match token.kind {
            TokenKind::ShiftLeft | TokenKind::ShiftRight => *span = peek,
            _ => return Ok(left_side_i),
        }
        let shift_node = SyntaxNode {
            kind: SyntaxNodeKind::ExpressionShift(token.kind),
            span: tree[left_side_i].span,
        };
        let shift_i = tree.insert(Insertion::ParentOf(left_side_i), shift_node);
        // left_side_i is no longer correct, as it has been displaced
        let right_side_i = parse_expr_add(tree, sources, span, Insertion::ChildOf(shift_i))?;
        tree[shift_i].span.as_mut().unwrap().end = tree[right_side_i].span.unwrap().end;
        left_side_i = shift_i;
    }
}

fn parse_expr_add(tree: &mut Tree,
                  sources: &SourceSet,
                  span: &mut Span,
                  insertion: Insertion)
                  -> Result<usize, Error> {
    let mut left_side_i = parse_expr_mul(tree, sources, span, insertion)?;
    loop {
        let mut peek = *span;
        let token = Token::next(&mut peek, sources)?;
        match token.kind {
            TokenKind::Plus | TokenKind::Minus => *span = peek,
            _ => return Ok(left_side_i),
        };
        let add_node = SyntaxNode {
            kind: SyntaxNodeKind::ExpressionAddition(token.kind),
            span: tree[left_side_i].span,
        };
        let add_i = tree.insert(Insertion::ParentOf(left_side_i), add_node);
        // left_side_i is no longer correct, as it has been displaced
        let right_side_i = parse_expr_mul(tree, sources, span, Insertion::ChildOf(add_i))?;
        tree[add_i].span.as_mut().unwrap().end = tree[right_side_i].span.unwrap().end;
        left_side_i = add_i;
    }
}

fn parse_expr_mul(tree: &mut Tree,
                  sources: &SourceSet,
                  span: &mut Span,
                  insertion: Insertion)
                  -> Result<usize, Error> {
    let mut left_side_i = parse_expr_cast(tree, sources, span, insertion)?;
    loop {
        let mut peek = *span;
        let token = Token::next(&mut peek, sources)?;
        match token.kind {
            TokenKind::Times | TokenKind::Division | TokenKind::Remainder => *span = peek,
            _ => return Ok(left_side_i),
        }
        let mul_node = SyntaxNode {
            kind: SyntaxNodeKind::ExpressionMultiplication(token.kind),
            span: tree[left_side_i].span,
        };
        let mul_i = tree.insert(Insertion::ParentOf(left_side_i), mul_node);
        // left_side_i is no longer correct, as it has been displaced
        let right_side_i = parse_expr_pow(tree, sources, span, Insertion::ChildOf(mul_i))?;
        tree[mul_i].span.as_mut().unwrap().end = tree[right_side_i].span.unwrap().end;
        left_side_i = mul_i;
    }
}

fn parse_expr_pow(tree: &mut Tree,
                  sources: &SourceSet,
                  span: &mut Span,
                  insertion: Insertion)
                  -> Result<usize, Error> {
    let mut left_side_i = parse_expr_cast(tree, sources, span, insertion)?;
    loop {
        let mut peek = *span;
        if Token::next(&mut peek, sources)?.kind != TokenKind::Power {
            return Ok(left_side_i);
        }
        *span = peek;
        let pow_node = SyntaxNode {
            kind: SyntaxNodeKind::ExpressionExponentiation(TokenKind::Power),
            span: tree[left_side_i].span,
        };
        let pow_i = tree.insert(Insertion::ParentOf(left_side_i), pow_node);
        // left_side_i is no longer correct, as it has been displaced
        let right_side_i = parse_expr_cast(tree, sources, span, Insertion::ChildOf(pow_i))?;
        tree[pow_i].span.as_mut().unwrap().end = tree[right_side_i].span.unwrap().end;
        left_side_i = pow_i;
    }
}

fn parse_expr_cast(tree: &mut Tree,
                   sources: &SourceSet,
                   span: &mut Span,
                   insertion: Insertion)
                   -> Result<usize, Error> {
    let mut left_side_i = parse_expr_unary(tree, sources, span, insertion)?;
    loop {
        let mut peek = *span;
        let token = Token::next(&mut peek, sources)?;
        match token.kind {
            TokenKind::Word if token.get(sources) == "as" => *span = peek,
            _ => return Ok(left_side_i),
        };
        let cast_node = SyntaxNode {
            kind: SyntaxNodeKind::ExpressionCast,
            span: tree[left_side_i].span,
        };
        let cast_i = tree.insert(Insertion::ParentOf(left_side_i), cast_node);
        // left_side_i is no longer correct, as it has been displaced
        let right_side_i = parse_expr(tree, sources, span, Insertion::ChildOf(cast_i))?;
        tree[cast_i].span.as_mut().unwrap().end = tree[right_side_i].span.unwrap().end;
        left_side_i = cast_i;
    }
}

fn parse_expr_unary(tree: &mut Tree,
                    sources: &SourceSet,
                    span: &mut Span,
                    insertion: Insertion)
                    -> Result<usize, Error> {
    let token = match parse_token_unary(sources, span) {
        Some(token) => token,
        _ => return parse_expr_group(tree, sources, span, insertion),
    };
    let unary_node = SyntaxNode {
        kind: SyntaxNodeKind::ExpressionUnary(token.kind),
        span: Some(token.span),
    };
    let unary_i = tree.insert(insertion, unary_node);
    let right_side_i = parse_expr_unary(tree, sources, span, Insertion::ChildOf(unary_i))?;
    tree[unary_i].span.as_mut().unwrap().end = tree[right_side_i].span.unwrap().end;
    Ok(unary_i)
}

fn parse_expr_group(tree: &mut Tree,
                    sources: &SourceSet,
                    span: &mut Span,
                    insertion: Insertion)
                    -> Result<usize, Error> {
    let mut peek = *span;
    let token = Token::next(&mut peek, sources)?;
    match token.kind {
        TokenKind::OpenParens => *span = peek,
        _ => return parse_expr_term(tree, sources, span, insertion),
    };
    let group_node = SyntaxNode {
        kind: SyntaxNodeKind::ExpressionGroup,
        span: Some(token.span),
    };
    let group_i = tree.insert(insertion, group_node);
    parse_expr(tree, sources, span, Insertion::ChildOf(group_i))?;
    let close = parse_token(sources, span, TokenKind::CloseParens, "')'")?;
    tree[group_i].span.as_mut().unwrap().end = close.span.end;
    parse_accessing(tree, sources, span, group_i)
}

fn parse_expr_term(tree: &mut Tree,
                   sources: &SourceSet,
                   span: &mut Span,
                   insertion: Insertion)
                   -> Result<usize, Error> {
    let token = Token::next(span, sources)?;
    let term_kind = match token.kind {
        TokenKind::Word => SyntaxNodeKind::Identifier(token.get(sources).to_string()),
        TokenKind::ZeroX => SyntaxNodeKind::DontCare,
        TokenKind::ZeroZ => SyntaxNodeKind::HighImpedance,
        TokenKind::Integer | TokenKind::IntegerXz => token.integer_literal(sources),
        TokenKind::Float => token.float_literal(sources),
        _ => return Err(token.err(sources, "expression")),
    };
    let term_node = SyntaxNode {
        kind: term_kind,
        span: Some(token.span),
    };
    let term_i = tree.insert(insertion, term_node);
    parse_accessing(tree, sources, span, term_i)
}

fn parse_accessing(tree: &mut Tree,
                   sources: &SourceSet,
                   span: &mut Span,
                   mut expr_i: usize)
                   -> Result<usize, Error> {
    loop {
        let mut peek = *span;
        let token = Token::next(&mut peek, sources)?;
        match token.kind {
            TokenKind::Point => expr_i = parse_expr_field(tree, sources, span, expr_i)?,
            TokenKind::OpenParens => expr_i = parse_expr_param(tree, sources, span, expr_i)?,
            TokenKind::OpenSquare => expr_i = parse_expr_index(tree, sources, span, expr_i)?,
            _ => return Ok(expr_i),
        }
    }
}

fn parse_expr_field(tree: &mut Tree,
                    sources: &SourceSet,
                    span: &mut Span,
                    expr_i: usize)
                    -> Result<usize, Error> {
    parse_token(sources, span, TokenKind::Point, "unreachable!")?;
    let mut access_node = SyntaxNode {
        kind: SyntaxNodeKind::ExpressionField,
        span: tree[expr_i].span,
    };
    let token = Token::next(span, sources)?;
    let ident = match token.kind {
        TokenKind::Word => token.get(sources).to_string(),
        _ => return Err(token.err(sources, "field identifier")),
    };
    let field_node = SyntaxNode {
        kind: SyntaxNodeKind::Identifier(ident),
        span: Some(token.span),
    };
    access_node.span.as_mut().unwrap().end = token.span.end;
    let access_i = tree.insert(Insertion::ParentOf(expr_i), access_node);
    // expr_i is no longer correct, as it has been displaced
    tree.insert(Insertion::ChildOf(access_i), field_node);
    Ok(access_i)
}

fn parse_expr_param(tree: &mut Tree,
                    sources: &SourceSet,
                    span: &mut Span,
                    expr_i: usize)
                    -> Result<usize, Error> {
    parse_token(sources, span, TokenKind::OpenParens, "unreachable!")?;
    let params_node = SyntaxNode {
        kind: SyntaxNodeKind::ExpressionParameters,
        span: tree[expr_i].span,
    };
    let params_i = tree.insert(Insertion::ParentOf(expr_i), params_node);
    // expr_i is no longer correct, as it has been displaced
    let mut need_comma = false;
    let mut insertion = Insertion::ChildOf(params_i);
    loop {
        let mut peek = *span;
        let token = Token::next(&mut peek, sources)?;
        match token.kind {
            TokenKind::Comma if need_comma => {
                *span = peek;
                need_comma = false;
                continue;
            }
            TokenKind::CloseParens => {
                *span = peek;
                tree[params_i].span.as_mut().unwrap().end = token.span.end;
                return Ok(params_i);
            }
            _ => {}
        }
        if need_comma {
            return Err(token.err(sources, "',' or ')'"));
        }
        let param_i = parse_expr(tree, sources, span, insertion)?;
        insertion = Insertion::SiblingOf(param_i);
        need_comma = true;
    }
}

fn parse_expr_index(tree: &mut Tree,
                    sources: &SourceSet,
                    span: &mut Span,
                    expr_i: usize)
                    -> Result<usize, Error> {
    parse_token(sources, span, TokenKind::OpenSquare, "unreachable!")?;
    let access_node = SyntaxNode {
        kind: SyntaxNodeKind::ExpressionIndex,
        span: tree[expr_i].span,
    };
    let access_i = tree.insert(Insertion::ParentOf(expr_i), access_node);
    // expr_i is no longer correct, as it has been displaced
    let mut peek = *span;
    let token = Token::next(&mut peek, sources)?;
    if token.kind == TokenKind::CloseSquare {
        *span = peek;
        tree[access_i].span.as_mut().unwrap().end = token.span.end;
        let index_node = SyntaxNode {
            kind: SyntaxNodeKind::Empty,
            span: None,
        };
        tree.insert(Insertion::ChildOf(access_i), index_node);
        return Ok(access_i);
    }
    parse_expr(tree, sources, span, Insertion::ChildOf(access_i))?;
    let close = parse_token(sources, span, TokenKind::CloseSquare, "']'")?;
    tree[access_i].span.as_mut().unwrap().end = close.span.end;
    Ok(access_i)
}

fn parse_token(sources: &SourceSet,
               span: &mut Span,
               required: TokenKind,
               expected: &str)
               -> Result<Token, Error> {
    let token = Token::next(span, sources)?;
    if token.kind != required {
        return Err(token.err(sources, expected));
    }
    Ok(token)
}

fn parse_token_unary(sources: &SourceSet, span: &mut Span) -> Option<Token> {
    let mut peek = *span;
    let token = match Token::next(&mut peek, sources) {
        Ok(token) => token,
        _ => return None,
    };
    match token.kind {
        TokenKind::Not | TokenKind::And | TokenKind::Or | TokenKind::Xor | TokenKind::Plus |
        TokenKind::Minus => {
            *span = peek;
            return Some(token);
        }
        TokenKind::Nand | TokenKind::Nor | TokenKind::Xnor => {}
        _ => return None,
    }
    // we cannot return !& !| or !^, we need to return just the !
    span.skip_whitespace(sources);
    let mut ret = Token {
        kind: TokenKind::Not,
        span: *span,
    };
    match span.get_whole(sources)[span.start..].char_indices().next() {
        Some((index, c)) if c == '!' => ret.span.end = span.start + index + c.len_utf8(),
        _ => unreachable!(),
    }
    span.column += 1;
    span.start = ret.span.end;
    Some(ret)
}

fn parse_dir(tree: &mut Tree,
             sources: &SourceSet,
             span: &mut Span,
             insertion: Insertion)
             -> Result<usize, Error> {
    let token = Token::next(span, sources)?;
    if token.kind != TokenKind::Word {
        return Err(token.err(sources, "direction ('in', 'out' or 'bidir')"));
    }
    let dir = match token.get(sources) {
        "in" => Direction::In,
        "out" => Direction::Out,
        "bidir" => Direction::Bidir,
        _ => return Err(token.err(sources, "direction ('in', 'out' or 'bidir')")),
    };
    let dir_node = SyntaxNode {
        kind: SyntaxNodeKind::Direction(dir),
        span: Some(token.span),
    };
    Ok(tree.insert(insertion, dir_node))
}
