// Copyright (C) 2016  University of Malta

// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#![allow(dead_code)]

use std::ops;
use std::usize;

#[derive(Copy, Clone, PartialEq, Debug)]
struct TreeNode<Node> {
    node: Node,
    child: usize,
    sibling: usize,
    garbage: usize,
}

#[derive(Clone, PartialEq, Debug)]
pub struct Tree<Node>(Vec<TreeNode<Node>>);

#[derive(Copy, Clone, PartialEq, Debug)]
pub enum Insertion {
    Root,
    ChildOf(usize),
    SiblingOf(usize),
    ParentOf(usize),
}

const DELETED: usize = usize::MAX;

impl<Node> Tree<Node> {
    pub fn new() -> Tree<Node> {
        Tree(vec![])
    }

    pub fn insert(&mut self, insertion: Insertion, node: Node) -> usize {
        let mut tree_node = TreeNode {
            node: node,
            child: 0,
            sibling: 0,
            garbage: 0,
        };
        match insertion {
            Insertion::Root => {
                if self.0.is_empty() {
                    self.0.push(tree_node);
                } else {
                    self.0[0] = tree_node;
                }
                0
            }
            Insertion::ChildOf(index) => {
                let mut child = self.0[index].child;
                if child == 0 {
                    let new_child = self.0.len();
                    self.0.push(tree_node);
                    self.0[index].child = new_child;
                    new_child
                } else {
                    while self.0[child].sibling != 0 {
                        child = self.0[child].sibling;
                    }
                    if self.0[child].garbage == DELETED {
                        // last child is marked for deletion and can be reused
                        self.0[child] = tree_node;
                        child
                    } else {
                        let new_child = self.0.len();
                        self.0.push(tree_node);
                        self.0[child].sibling = new_child;
                        new_child
                    }
                }
            }
            Insertion::SiblingOf(index) => {
                let sibling = self.0[index].sibling;
                if self.0[index].garbage == DELETED {
                    // original sibling is marked for deletion and can be reused
                    tree_node.sibling = sibling;
                    self.0[index] = tree_node;
                    index
                } else if sibling != 0 && self.0[sibling].garbage == DELETED {
                    // first sibling is marked for deletion and can be reused
                    tree_node.sibling = self.0[sibling].sibling;
                    self.0[sibling] = tree_node;
                    sibling
                } else {
                    // insert between original sibling and its first sibling
                    let new_sibling = self.0.len();
                    tree_node.sibling = sibling;
                    self.0.push(tree_node);
                    self.0[index].sibling = new_sibling;
                    new_sibling
                }
            }
            Insertion::ParentOf(index) => {
                let new_index = self.0.len();
                tree_node.child = new_index;
                tree_node.sibling = self.0[index].sibling;
                self.0.push(tree_node);
                self.0[index].sibling = 0;
                self.0.swap(index, new_index);
                index
            }
        }
    }

    // children of removed 0 are removed as well, but siblings are not removed
    pub fn remove(&mut self, index: usize) {
        self.0[index].garbage = DELETED;
    }

    fn mark_children(&mut self, index: usize) {
        self.0[index].garbage = 0;
        let mut first_child = self.0[index].child;
        while first_child != 0 && self.0[first_child].garbage == DELETED {
            first_child = self.0[first_child].sibling;
        }
        self.0[index].child = first_child;
        if first_child == 0 {
            return;
        }
        self.mark_children(first_child);
        loop {
            let mut next_child = self.0[first_child].sibling;
            while next_child != 0 && self.0[next_child].garbage == DELETED {
                next_child = self.0[next_child].sibling;
            }
            self.0[first_child].sibling = next_child;
            if next_child == 0 {
                return;
            }
            self.mark_children(next_child);
            first_child = next_child;
        }
    }

    pub fn collect_garbage(&mut self) {
        if self.0.is_empty() {
            return;
        }
        if self.0[0].garbage == DELETED {
            self.0.clear();
            return;
        }
        let len = self.0.len();
        // Initially, garbage = DELETED for all deleted 0, != DELETED otherwise.
        for i in 1..len {
            if self.0[i].garbage != DELETED {
                self.0[i].garbage = 1;
            }
        }
        // Now garbage = DELETED for all deleted 0, 1 otherwise.
        self.mark_children(0);
        // Now garbage = DELETED for deleted 0, 1 for unreachable 0, 0 for reachable 0.
        // Also, child and sibling indices have now bypassed all unreachable/deleted 0.
        let mut del_count = 0;
        for i in 1..len {
            if self.0[i].garbage == 0 {
                self.0[i].garbage = del_count;
            } else {
                self.0[i].garbage = DELETED;
                del_count += 1;
            }
        }
        // Now garbage = DELETED for deleted or unreachable 0, del_count otherwise,
        // where del_count is the count of previous 0 that are to be removed.
        // That is indices of these 0 will be decreased by del_count.
        for i in 0..len {
            if self.0[i].garbage == DELETED {
                continue;
            }
            let child = self.0[i].child;
            self.0[i].child -= self.0[child].garbage;
            let sibling = self.0[i].sibling;
            self.0[i].sibling -= self.0[sibling].garbage;
        }
        // Now we just remove the garbage.
        self.0.retain(|node| node.garbage != DELETED);
    }

    pub fn root(&self) -> Option<usize> {
        if self.0.is_empty() || self.0[0].garbage == DELETED {
            None
        } else {
            Some(0)
        }
    }

    pub fn child(&self, index: usize) -> Option<usize> {
        let mut child = self.0[index].child;
        while child > 0 && self.0[child].garbage == DELETED {
            child = self.0[child].sibling;
        }
        if child == 0 { None } else { Some(child) }
    }

    pub fn sibling(&self, index: usize) -> Option<usize> {
        let mut sibling = self.0[index].sibling;
        while sibling > 0 && self.0[sibling].garbage == DELETED {
            sibling = self.0[sibling].sibling;
        }
        if sibling == 0 { None } else { Some(sibling) }
    }
}

impl<Node> ops::Index<usize> for Tree<Node> {
    type Output = Node;
    fn index(&self, index: usize) -> &Node {
        &self.0[index].node
    }
}

impl<Node> ops::IndexMut<usize> for Tree<Node> {
    fn index_mut(&mut self, index: usize) -> &mut Node {
        &mut self.0[index].node
    }
}
